#include "Behaviour.h"

#include <iostream>
#include <Eigen/SVD>

/**
   @brief Constructor
   @param ReferenceTransformhandler - the Transformhandler storing the current configuration
   @param OwningTaskLevelController - TaskLevelController, from which the behavior is called
   @param name - name of the milestone (for debug)
*/

using namespace std;

Behaviour::Behaviour(Transformhandler* ReferenceTransformhandler,
		TaskLevelController* OwningTaskLevelController)
{
	myTaskLevelController = OwningTaskLevelController;
	BehaviorTransformhandler =  ReferenceTransformhandler;
	myTorques.setZero();
	NumberOfDOF = myTaskLevelController->getNumberOfDOF();
}

/**
   @brief Desctructor.
*/
Behaviour::~Behaviour() {}

const Eigen::MatrixXd Behaviour::getTaskJacobian()
{
	  return TaskJacobian;
}

const Eigen::MatrixXd Behaviour::getResultingNullspace()
{
	return NullspaceForNextBehavior;
}

bool Behaviour::getActive()
{
	return BehaviorActive;
}

/** @brief Calculates the jacobian for all the joints for the operational point at the origin
 * 			of link with index JacobianOperationPointIndex.
 * 			if JacobianOperationPointIndex == (NumberOfJoints - 1), jacobian of the end effector is calculated.
 * 			if JacobianOperationPointIndex == 0, jacobian of the base (2 DOF) is calculated
 *  @param JacobianOperationPointIndex - must be a number between 0 and (NumberOfJoints - 1)
 *  @param mask - sets, which rows of theJacobian should be computed (depends on the task)
 * */
Eigen::MatrixXd Behaviour::calculateOperationPointJacobian(uint JacobianOperationPointIndex, MaskArray mask)
{
  //Getting data about configuration
	vector<btVector3> rotAxes;
	vector<btVector3> rotOrigins;
	rotAxes = BehaviorTransformhandler->getAxes();

	rotOrigins = BehaviorTransformhandler->getOrigins();

	int NumberOfJacobianRows = 0;
	for(int j=0; j<6; j++)
		if (mask[j])
			NumberOfJacobianRows++;
	//Temporary local variables
	Eigen::MatrixXd Jac_buf;
	Jac_buf.setZero(NumberOfJacobianRows,NumberOfDOF);
	if (JacobianOperationPointIndex > rotOrigins.size())
	{
		cout<<"Behaviour::calculateOperationPointJacobian: ERROR: A jacobian for a non-existing joint requested."<<endl;
		return Jac_buf;
	}
	Eigen::VectorXd Jac_column;
	Jac_column.resize(6);
	Eigen::Vector3d Jv, endOrigin, prevOrigin, prevZ;
	Jv.setZero();
	Jac_column.setZero();
	endOrigin.setZero();
	prevOrigin.setZero();
	prevZ.setZero();

	//Position of operation point
	endOrigin = Eigen::Vector3d(rotOrigins.at(JacobianOperationPointIndex).m_floats);
	//Arm joints
        for (uint i = 0; i <= JacobianOperationPointIndex; i++)
        {
                // Converting axes and origins from btVector3 to Eigen::Vector3d
                prevZ = Eigen::Vector3d(rotAxes[i].m_floats);
                prevOrigin = Eigen::Vector3d(rotOrigins[i].m_floats);
                // Revolute joint jacobian:
                // J_vi = z_i-1 x (o_n - o_n-1)
                Jv = prevZ.cross(endOrigin - prevOrigin);

                Jac_column.topRows(3) = Jv;
                Jac_column.bottomRows(3) = prevZ;
                int k=0;
                for(int j=0; j<6; j++)
                        if (mask[j])
                        {
                                Jac_buf.col(i)[k] = Jac_column[j];
                                k++;
                        }
        }
//        cout<<Jac_buf<<endl<<endl;
	return Jac_buf;

}

/** @brief	Computes the task Jacobian of an Operation point.
 * 			Uses OperationalDOFMask to choose, which rows of Jacobian to compute.
*/
Eigen::MatrixXd Behaviour::calculateOperationPointJacobian(uint JacobianOperationPointIndex)
{
	return calculateOperationPointJacobian(JacobianOperationPointIndex, OperationalDOFMask);
}

/** @brief	Function transforms reference acceleration of operational point
 * 			to configuration-space torque vector. For behaviors working in configuration space it should not be used.
 * 			The TaskJacobian and Lambda_p must be calculated before this function is called.
*/
void Behaviour::TransformRefAccelerationToTorques()
{
  if (!BehaviorActive)
	{
		myTorques.setZero();
		return;
	}
	Eigen::MatrixXd N_prev;
	N_prev = myTaskLevelController->getNullSpaceMatrix();
	Eigen::MatrixXd J_p= TaskJacobian * N_prev;
	/*TODO: Compute Coriolis and gravitation parts with Newton-Euler algorithm and
		add their contribution to the force*/
	Eigen::VectorXd Force = Lambda_p * TaskReferenceAcceleration;
	myTorques = J_p.transpose() * Force;
}

/** @brief	Function asks Task-level controller for the nullspace of previous tasks,
 * 			and uses it and TaskJacobian to compute The resulting nullspace.
 * 			The TaskJacobian must be computed before the function is called.
*/
void Behaviour::calculateNextTaskNullspace()
{
 	const int NumberOfJacobianRows = TaskJacobian.rows();
	Eigen::MatrixXd N_prev;
	N_prev = myTaskLevelController->getNullSpaceMatrix();

	if (!BehaviorActive || TaskJacobian.cols()!=NumberOfDOF || NumberOfJacobianRows==0)
	{
		TaskJacobian.setZero();
		NullspaceForNextBehavior = N_prev;
		return;
	}

//************TASK LAMBDA CALCULATION*************//
	Eigen::MatrixXd J_p;
	J_p= TaskJacobian * N_prev;
	Eigen::MatrixXd H_inverse = myTaskLevelController->getInertiaMatrixInverse();
	Eigen::MatrixXd Sigma_inverse, Lambda_P_Inverse;
	Lambda_P_Inverse = J_p * H_inverse * J_p.transpose();
	//compute SVD decomposition to check for singular values
	Eigen::JacobiSVD<Eigen::MatrixXd> svd(Lambda_P_Inverse, Eigen::ComputeThinU | Eigen::ComputeThinV);
	Sigma_inverse.setZero(NumberOfJacobianRows,NumberOfJacobianRows);
	for (int i = 0; i<NumberOfJacobianRows; i++)
		if (svd.singularValues()[i] > 0.001)
			Sigma_inverse(i,i) = 1 / svd.singularValues()[i];
	Lambda_p = svd.matrixU() * Sigma_inverse * svd.matrixU().transpose();
//*********NULLSPACE CALCULATION *******//
	Eigen::MatrixXd J_gen_inv;
	J_gen_inv = H_inverse * J_p.transpose() * Lambda_p;

	NullspaceForNextBehavior = (N_prev - J_gen_inv * J_p);
}


/**
   @brief Inheriting classes should do the work in this function.
*/
void Behaviour::calculateReferenceAcceleration() { }



/**
   @brief Calculates the joint torques and then returns them.
   	   	   For tasks, acting in configuration space should be overriden
   	   	    	not to call TransformRefAccelerationToTorques().
   NOTE: torques calculation has no notion of time limits
*/
const torqueVector& Behaviour::getTorque()
{
  calculateReferenceAcceleration();
  calculateNextTaskNullspace();
  TransformRefAccelerationToTorques();
  return myTorques;
}


