#ifndef __BEHAVIOUR_H__
#define __BEHAVIOUR_H__


#include <btBulletDynamicsCommon.h>
#include <Eigen/Core>
#include <Eigen/Geometry>

#include "../matrixDefinition.h"
#include "../Transformhandler.h"
#include "../TaskLevelController.h"


using namespace boost::posix_time;

#define DEFAULT_BEHAVIOR_MAX_FORCE 30.0

class TaskLevelController;

typedef bool MaskArray[6];

/**
   @brief Behaviour superclass used by TaskLevelControllers.
   A behavior calculates desired movement of the operating point,
   takes the nullspace matrix of all higher-priority tasks,
   provides the torque vector consistent with higher-priority tasks
   and updates the resulting nullspace matrix for next behaviors.

   Refer to  L. Sentis and O. Khatib, "Synthesis of whole-body behaviors through hierarchical control of
		behavioral primitives" for details.
*/
class Behaviour {
 public:
  Behaviour(Transformhandler* ReferenceTransformhandler,
		  TaskLevelController* OwningTaskLevelController);
  ~Behaviour();
  virtual const torqueVector& getTorque();
  const Eigen::MatrixXd getTaskJacobian();
  const Eigen::MatrixXd getResultingNullspace();
  bool getActive();
  EIGEN_MAKE_ALIGNED_OPERATOR_NEW //Aligning bytes in matrices (libEigen tricks)

  enum OperationSpaceFreeMotions{
	  MOVE_X,
	  MOVE_Y,
	  MOVE_Z,
	  ROTATE_X,
	  ROTATE_Y,
	  ROTATE_Z
  };

 protected:
  virtual void calculateReferenceAcceleration() = 0;
  Eigen::MatrixXd calculateOperationPointJacobian(uint JacobianOperationPointIndex);
  Eigen::MatrixXd calculateOperationPointJacobian(uint JacobianOperationPointIndex, MaskArray mask);
  void calculateNextTaskNullspace();
  void TransformRefAccelerationToTorques();
  inline int signof(double a) { return (a != 0) ? (a<0 ? -1 : 1) : 0; }

  Eigen::MatrixXd TaskJacobian;
  /**@brief This nullspace combines the nullspace from preceding tasks
   *  and the nullspace of the current task. It's computed in calculateNextTaskNullspace().  */
  Eigen::MatrixXd NullspaceForNextBehavior;
  /**@brief Vector of torques consistent with nullspace of previous tasks  */
  torqueVector myTorques;
  Eigen::VectorXd myForce;
  /**@brief Vector of desired Acceleration of operational point in operational space.
   * 		for behaviors working in configuration-space (e.g., joint angle limitation)  it's not used. */
  Eigen::VectorXd TaskReferenceAcceleration;
  /**@brief Operational-space inertia matrix consistent with all preceding tasks.
   * 		Used for Task Nullspace calculation and transformation of TaskReferenceAcceleration
   * 		to desired torque vector.
   * 		It's computed in calculateNextTaskNullspace().*/
  Eigen::MatrixXd Lambda_p;

  /** @brief Index of a link, whose coordinate frame origin is an
   * 	operational point for the behavior. Should be a number
   * 	between 0 and (NumberOfJoints - 1)*/
  int OperationPointIndex;
  /** @brief Array of bools, specifying which free motions of operation point
   * 	should be considered by the behavior (1 - active; 0 - disregard)*/
  MaskArray OperationalDOFMask;

  bool BehaviorActive;
  int NumberOfDOF;
  TaskLevelController* myTaskLevelController;
  Transformhandler* BehaviorTransformhandler;
  std::string * milestone_name;
};

#endif /*__BEHAVIOUR_H__*/
