#include "KinematicBehaviour.h"

#include <stdlib.h>

KinematicBehaviour::KinematicBehaviour(Transformhandler* ReferenceTransformhandler,
		TaskLevelController* OwningTaskLevelController,double limit_zone)
  : Behaviour(ReferenceTransformhandler,OwningTaskLevelController)
{
	BehaviorActive = true;
  CloseToLimitZone = limit_zone;
  myMultiplier = 1.0;
  myTaskLevelController->getRobot()->getTransformhandler()->getJointLimits(myJointLimits,jointIsLimited);
  OperationPointIndex = BehaviorTransformhandler->getOrigins().size()-1;
  TaskReferenceAcceleration.setZero(3);
  OperationalDOFMask[MOVE_X] = true;
  OperationalDOFMask[MOVE_Y] = true;
  OperationalDOFMask[MOVE_Z] = true;
  OperationalDOFMask[ROTATE_X] = false;
  OperationalDOFMask[ROTATE_Y] = false;
  OperationalDOFMask[ROTATE_Z] = false;
}

KinematicBehaviour::~KinematicBehaviour()
{}


/**
   @brief Output force scaling. Default scale = 1.0
 */
void KinematicBehaviour::setMultiplier(double x) {
  myMultiplier = x;
}

/**
   @brief Calculates the joint torques and then returns them.

   NOTE: torques calculation has no notion of time limits
*/
const torqueVector& KinematicBehaviour::getTorque()
{
  calculateReferenceAcceleration();
  calculateNextTaskNullspace();
  return myTorques;
}

/* Kinematic behavior works in configuration space and it has the highest priority.
 * if a joint is close to its limit, then the behavior is active and tries to move
 * the joint in an opposite direction. The lower-priority behaviors cannot affect
 * these joints any more.
*/
void KinematicBehaviour::calculateReferenceAcceleration()
{
  Eigen::MatrixXd TaskJacobianBuf = calculateOperationPointJacobian(OperationPointIndex);
  TaskJacobian.setZero(TaskJacobianBuf.rows(),TaskJacobianBuf.cols());

  //Getting data about configuration
  cVector q_current = BehaviorTransformhandler->getConfiguration();

  torqueVector Desired_q_speed;
  Desired_q_speed.setZero();
  torqueVector Actual_q_speed = myTaskLevelController->getSpeedVector();
  torqueVector Desired_q_acceleration;
  Desired_q_acceleration.setZero();

  double Kv = 2;
  double Kp = 0.5;
  for(int i=0;i<jointIsLimited.rows();i++)
  {
    if (jointIsLimited[i])
    {
      if (q_current[i]>myJointLimits(1,i)-CloseToLimitZone)
      {
        Desired_q_speed[i] = -Kp * (1 + (q_current[i]-myJointLimits(1,i))/CloseToLimitZone);
        if (fabs(Desired_q_speed[i]) > 0.44)
          Desired_q_speed[i] = 0.44 * signof(Desired_q_speed[i]);
        Desired_q_acceleration[i] = -Kv * (Actual_q_speed[i] - Desired_q_speed[i]);
        if (fabs(Desired_q_acceleration[i]) > 1.75)
          Desired_q_acceleration[i] = 1.75 * signof(Desired_q_speed[i]);
        TaskJacobian.col(i) = TaskJacobianBuf.col(i);
      }
      if (q_current[i]<myJointLimits(0,i)+CloseToLimitZone)
      {
        Desired_q_speed[i] = Kp * (1 + (-q_current[i]+myJointLimits(0,i))/CloseToLimitZone);
        if (fabs(Desired_q_speed[i]) > 0.44)
          Desired_q_speed[i] = 0.44 * signof(Desired_q_speed[i]);
        Desired_q_acceleration[i] = -Kv * (Actual_q_speed[i] - Desired_q_speed[i]);
        if (fabs(Desired_q_acceleration[i]) > 1.75)
          Desired_q_acceleration[i] = 1.75 * signof(Desired_q_speed[i]);
        TaskJacobian.col(i) = TaskJacobianBuf.col(i);
      }
    }
  }

  //TODO: Compute Coriolis and gravitation parts with Newton-Euler algorithm and
  // 			add their contribution to the torques
  torqueVector DesiredTorques = myTaskLevelController->getInertiaMatrix() * Desired_q_acceleration;

  Eigen::MatrixXd N_prev = myTaskLevelController->getNullSpaceMatrix();
  myTorques = N_prev.transpose() * DesiredTorques;

  BehaviorActive = DesiredTorques.norm() > 0;

}
