/*
 * Obstacle.h
 *
 *  Created on: 02.05.2012
 *      Author: sergey
 */

#ifndef OBSTACLE_H_
#define OBSTACLE_H_

#include <btBulletDynamicsCommon.h>
#include "matrixDefinition.h"
#include "Cell.h"
#include <map>

class Obstacle
{
public:
  enum {
    CUBE,
    SPHERE,
    OBST_TYPE_END
  };

  Obstacle ()
  {
    transform.setIdentity();
    }
  Obstacle (btVector3 cen, btQuaternion rot = btQuaternion(btVector3(0,0,1),0))
  {
    transform.setOrigin(cen);
    transform.setRotation(rot);
  }

  Obstacle (btTransform trans)
  {
    transform = trans;
  }
  btTransform getTransform() {return transform;}

  void setTransform(btTransform newTransform) {
    transform = newTransform;
    findBoundingCube();}
  void setPosition(btVector3 newPos) {
    transform.setOrigin(newPos);
    findBoundingCube();}
  void setRotation(btQuaternion newRot) {
    transform.setRotation(newRot);
    findBoundingCube();}

  void getBoundingCube(btVector3& min, btVector3& max) {
    min = boundingCubeMinPoint;
    max = boundingCubeMaxPoint;
  }
  virtual bool pointIsInsideObstacle(btVector3 point) = 0;
  map<long int, Cell*> occupied_cells;

  uint getType() {return type;}
protected:
  virtual void findBoundingCube() = 0;

  btVector3 boundingCubeMinPoint, boundingCubeMaxPoint;
  btTransform transform;
  uint type;
};





class CubicObstacle : public Obstacle {
public:
  CubicObstacle() : Obstacle()
    {dimensions.setZero(); type = Obstacle::CUBE;}

  CubicObstacle(btVector3 dim, btVector3 cent, btQuaternion rot = btQuaternion(btVector3(0,0,1),0)) :
    Obstacle(cent,rot), dimensions(dim) {
    type = Obstacle::CUBE;
    findBoundingCube();}

  CubicObstacle(btVector3 dim, btTransform trans) :
    Obstacle(trans), dimensions(dim) {
    type = Obstacle::CUBE;
    findBoundingCube();}


  virtual ~CubicObstacle() {};

  btVector3 getDim() {return dimensions;}
  void setDim(btVector3 newDim) {
    dimensions = newDim;
    findBoundingCube(); }

  bool pointIsInsideObstacle (btVector3 point)
  {
    btVector3 obstToPoint = transform.inverse() * point;
    return (obstToPoint[0] <= dimensions[0]/2 && obstToPoint[0] >= -dimensions[0]/2 &&
        obstToPoint[1] <= dimensions[1]/2 && obstToPoint[1] >= -dimensions[1]/2 &&
        obstToPoint[2] <= dimensions[2]/2 && obstToPoint[2] >= -dimensions[2]/2 );
  }
private:

  btVector3 dimensions;
  void findBoundingCube()
  {
    btVector3 points[8];
    points[0] = dimensions/2;
    points[1] = dimensions/2;
    points[1][0] *= -1;
    points[2] = dimensions/2;
    points[2][1] *= -1;
    points[3] = dimensions/2;
    points[3][2] *= -1;
    points[4] = -dimensions/2;
    points[5] = -dimensions/2;
    points[5][0] *= -1;
    points[6] = -dimensions/2;
    points[6][1] *= -1;
    points[7] = -dimensions/2;
    points[7][2] *= -1;

    boundingCubeMaxPoint = btVector3(MIN_X_RANGE,MIN_Y_RANGE,MIN_Z_RANGE);
    boundingCubeMinPoint = btVector3(MAX_X_RANGE,MAX_Y_RANGE,MAX_Z_RANGE);

    for (int i=0;i<8;i++)
    {
      points[i] = transform * points[i];
      if(points[i][0]<boundingCubeMinPoint[0])
       boundingCubeMinPoint[0] = points[i][0];
      if(points[i][1]<boundingCubeMinPoint[1])
       boundingCubeMinPoint[1] = points[i][1];
      if(points[i][2]<boundingCubeMinPoint[2])
       boundingCubeMinPoint[2] = points[i][2];

      if(points[i][0]>boundingCubeMaxPoint[0])
        boundingCubeMaxPoint[0] = points[i][0];
      if(points[i][1]>boundingCubeMaxPoint[1])
        boundingCubeMaxPoint[1] = points[i][1];
      if(points[i][2]>boundingCubeMaxPoint[2])
        boundingCubeMaxPoint[2] = points[i][2];
    }
  }

};




class SphericObstacle : public Obstacle
{
public:
  SphericObstacle() : Obstacle(),radius(0) {
    type = Obstacle::SPHERE;
    findBoundingCube(); }
  SphericObstacle(double rad, btVector3 cent, btQuaternion rot = btQuaternion(btVector3(0,0,1),0)) :
    Obstacle(cent,rot), radius(rad) {
    type = Obstacle::SPHERE;
    findBoundingCube(); }

  SphericObstacle(double rad, btTransform trans) :
    Obstacle(trans),radius(rad) {
    type = Obstacle::SPHERE;
    findBoundingCube(); }
  virtual ~SphericObstacle() {}
  double getRadius () {return radius;}

    bool pointIsInsideObstacle (btVector3 point)
    {
      btVector3 obstToPoint = transform.inverse() * point;
      return (obstToPoint.length() <= radius );
    }
private:
    double radius;

    void findBoundingCube()
    {
      btVector3 a(1,1,1);
      boundingCubeMaxPoint = transform.getOrigin() + a*radius;
      boundingCubeMinPoint = transform.getOrigin() - a*radius;
    }
};






#endif /* OBSTACLE_H_ */
