/*
 * ObstaclesHandler.cpp
 *
 *  Created on: 02.05.2012
 *      Author: sergey
 */

#include "ObstaclesHandler.h"
#include <visualization_msgs/MarkerArray.h>

ObstaclesHandler::ObstaclesHandler(ros::NodeHandle* n, Planner* planner)
{
  node_handler = n;
  myCells = planner->getCellArray();
  myPlanner = planner;
  obstacle_publisher = n->advertise<visualization_msgs::MarkerArray>("visualization_marker_array", 0);


  manipulator_base = new CubicObstacle();
  double r = 0.06;
  double x,y,z;
  while (z<0.2)
  {
    while (y<r)
    {
      while (x<r)
      {
        if (x*x + y*y <= r*r)
        {
          long int ID = FormIDfromPos(btVector3(x,y,z));
          if(myCells->count(ID) != 0)
            manipulator_base->occupied_cells.insert(pair<long int, Cell*>(ID,&(myCells->find(ID)->second)));
          else
            cout<<"base: cant find cell "<<ID<<"("<<x<<" "<<y<<" "<<z<<")"<<endl;
        }
        x += CELL_GRID_STEP;
      }
      y+=CELL_GRID_STEP;
      x=-r;
    }
    z+= CELL_GRID_STEP;
    y=-r;
  }

  myObstacles.push_back(manipulator_base);


}

ObstaclesHandler::~ObstaclesHandler()
{
  //updateObstaclesRviz(true);
  //obstacle_publisher.shutdown();

  for (uint i =0; i<myObstacles.size();i++ )
    delete myObstacles.at(i);
  cout<<"ObstaclesHandler: END"<<endl;
}

Obstacle* ObstaclesHandler::addObstacle(uint type, btVector3 dim, btVector3 cent, btQuaternion rot)
{
  Obstacle* newObstacle;
  switch (type){
    case (Obstacle::CUBE):
        newObstacle = new CubicObstacle(dim,cent,rot);
    break;
    case (Obstacle::SPHERE):
        newObstacle = new SphericObstacle(dim[0],cent,rot);
    break;
    default:
      return NULL;
  }

  myObstacles.push_back(newObstacle);
  if (newObstacle == NULL)
    return newObstacle;
  ros::Time start_time = ros::Time::now();

  btVector3 cubeMin, cubeMax;
  newObstacle->getBoundingCube(cubeMin, cubeMax);

//  cout<<"Cube obstacle "<<myObstacles.size()
//      <<" from ("<<cubeMin.getX()<<" "<<cubeMin.getY()<<" "<<cubeMin.getZ()<<") to ("
//      <<cubeMax.getX()<<" "<<cubeMax.getY()<<" "<<cubeMax.getZ()<<")"<<endl;
  cubeMin[0] = CELL_GRID_STEP * (int)(cubeMin[0]/CELL_GRID_STEP -1);
  cubeMin[1] = CELL_GRID_STEP * (int)(cubeMin[1]/CELL_GRID_STEP -1);
  cubeMin[2] = CELL_GRID_STEP * (int)(cubeMin[2]/CELL_GRID_STEP -1);

  cubeMax[0] = CELL_GRID_STEP * (int)(cubeMax[0]/CELL_GRID_STEP +1);
  cubeMax[1] = CELL_GRID_STEP * (int)(cubeMax[1]/CELL_GRID_STEP +1);
  cubeMax[2] = CELL_GRID_STEP * (int)(cubeMax[2]/CELL_GRID_STEP +1);

  double x = cubeMin[0],y=cubeMin[1],z=cubeMin[2];
  long int ID;
  Cell* cellWithinBoundCube;
  while (z<=cubeMax[2])
  {
    while(y<=cubeMax[1])
    {
      while(x<=cubeMax[0])
      {
        ID = FormIDfromPos(btVector3(x,y,z));
        if(myCells->count(ID) != 0)
        {
          cellWithinBoundCube = &(myCells->find(ID)->second);

           if (newObstacle->pointIsInsideObstacle(cellWithinBoundCube->getCenter()) )
            //if the cell center is within obstacle
          {
            newObstacle->occupied_cells.insert(pair<long int, Cell*>(cellWithinBoundCube->getId(),cellWithinBoundCube));
            vector< pair<long int,uint> >::iterator it;
            Vertex n;
            for (it =cellWithinBoundCube->getConfigList()->begin(); it!=cellWithinBoundCube->getConfigList()->end(); it++)
            {
              n = myPlanner->getID2VertexMap()->find((*it).first)->second;
              myPlanner->roadmap[n].config[(*it).second].invalid++;
            } //for
          } //if
        }//if
        else
          if (x>MIN_X_RANGE && x<MAX_X_RANGE && y>MIN_Y_RANGE && y<MAX_Y_RANGE && z>MIN_Z_RANGE && z<MAX_Z_RANGE)
            cout<<"WARNING: ObstaclesHandler::addObstacle: Cell "<<ID<<"(center = "<<x<<" "<<y<<" "<<z<<") not found!!!"<<endl;
        x+=CELL_GRID_STEP;
      } //while x
      y+=CELL_GRID_STEP;
      x=cubeMin[0];
    } //while y
    z+=CELL_GRID_STEP;
    y=cubeMin[1];
  } //while z
//  cout<<"Cube obstacle "<<myObstacles.size()
//      <<" from ("<<cubeMin.getX()<<" "<<cubeMin.getY()<<" "<<cubeMin.getZ()<<") to ("
//      <<cubeMax.getX()<<" "<<cubeMax.getY()<<" "<<cubeMax.getZ()<<")"<<endl;

  updateObstaclesRviz();

  cout<<"ObstaclesHandler::addObstacle: Function duration "<<ros::Time::now() - start_time<<" sec"<<endl;
  return newObstacle;
}

Obstacle* ObstaclesHandler::addObstacle(uint type,btVector3 dim,  btTransform trans)
{
  return addObstacle(type, dim,trans.getOrigin(),trans.getRotation());
}

bool ObstaclesHandler::moveObstacle(Obstacle* obst, btVector3 newCenter, btQuaternion newRot)
{
  ros::Time start_time = ros::Time::now();

  vector<Obstacle*>::iterator it;
  for (it=myObstacles.begin(); it!= myObstacles.end(); it++)
    if((*it) == obst)
    {
      (*it)->setPosition(newCenter);
      (*it)->setRotation(newRot);
      updateObstaclesRviz();

      btVector3 cubeMin, cubeMax;
      obst->getBoundingCube(cubeMin, cubeMax);

    //  cout<<"Cube obstacle "<<myObstacles.size()
    //      <<" from ("<<cubeMin.getX()<<" "<<cubeMin.getY()<<" "<<cubeMin.getZ()<<") to ("
    //      <<cubeMax.getX()<<" "<<cubeMax.getY()<<" "<<cubeMax.getZ()<<")"<<endl;
      cubeMin[0] = CELL_GRID_STEP * (int)(cubeMin[0]/CELL_GRID_STEP -1);
      cubeMin[1] = CELL_GRID_STEP * (int)(cubeMin[1]/CELL_GRID_STEP -1);
      cubeMin[2] = CELL_GRID_STEP * (int)(cubeMin[2]/CELL_GRID_STEP -1);

      cubeMax[0] = CELL_GRID_STEP * (int)(cubeMax[0]/CELL_GRID_STEP +1);
      cubeMax[1] = CELL_GRID_STEP * (int)(cubeMax[1]/CELL_GRID_STEP +1);
      cubeMax[2] = CELL_GRID_STEP * (int)(cubeMax[2]/CELL_GRID_STEP +1);

      double x = cubeMin[0],y=cubeMin[1],z=cubeMin[2];
      long int ID;
      Cell* cellWithinBoundCube;
      map<long int, Cell*> cellsWitinObstAfterMove;

      //form a list of cells occupied by the obstacle after move
      while (z<=cubeMax[2])
      {
        while(y<=cubeMax[1])
        {
          while(x<=cubeMax[0])
          {
            ID = FormIDfromPos(btVector3(x,y,z));
            if(myCells->count(ID) != 0)
            {
              cellWithinBoundCube = &(myCells->find(ID)->second);

              if (obst->pointIsInsideObstacle(cellWithinBoundCube->getCenter()) )
                //if the cell center is within obstacle add it to the list
                cellsWitinObstAfterMove.insert(pair<long int, Cell*>(cellWithinBoundCube->getId(),cellWithinBoundCube));
            }//if
            else
              if (x>MIN_X_RANGE && x<MAX_X_RANGE && y>MIN_Y_RANGE && y<MAX_Y_RANGE && z>MIN_Z_RANGE && z<MAX_Z_RANGE)
                cout<<"WARNING: ObstaclesHandler::moveObstacle: Cell "<<ID<<"(center = "<<x<<" "<<y<<" "<<z<<") not found!!!"<<endl;
            x+=CELL_GRID_STEP;
          } //while x
          y+=CELL_GRID_STEP;
          x=cubeMin[0];
        } //while y
        z+=CELL_GRID_STEP;
        y=cubeMin[1];
      } //while z

      //now compare two lists
      map<long int, Cell*>::iterator cell_iter = cellsWitinObstAfterMove.begin();

      //check if old cells are in the new set.
      //if yes - do nothing. if no - validate colliding configurations and delete the cell
      for (cell_iter=obst->occupied_cells.begin(); cell_iter != obst->occupied_cells.end(); cell_iter++)
      {
       // cout<<"Cell "<<(*cell_iter).first<<" pos: "<<(*cell_iter).second->getCenter()[0]<<" "<<(*cell_iter).second->getCenter()[1]<<" "<<(*cell_iter).second->getCenter()[2]<<endl;

        if (cellsWitinObstAfterMove.count((*cell_iter).first) == 0)
        {
          Cell* cellToDelete = (*cell_iter).second;
          vector< pair<long int,uint> >::iterator config_it;
          Vertex n;
          for (config_it =cellToDelete->getConfigList()->begin(); config_it!=cellToDelete->getConfigList()->end(); config_it++)
          {
            n = myPlanner->getID2VertexMap()->find((*config_it).first)->second;
            myPlanner->roadmap[n].config[(*config_it).second].invalid--;
          } //for
          obst->occupied_cells.erase(cell_iter);
          cell_iter--;

        }
      }

      //check if new cells are in the old set.
      //if yes - do nothing. if no - invalidate colliding configurations, check if the current path is obstructed, add the cell to list
      list<Vertex> currentPath = myPlanner->getCurrentPathNodes();
      list<Vertex>::iterator path_iter;
      for (cell_iter=cellsWitinObstAfterMove.begin(); cell_iter != cellsWitinObstAfterMove.end(); cell_iter++)
      {
        //cout<<"Cell "<<(*cell_iter).first<<" pos: "<<(*cell_iter).second->getCenter()[0]<<" "<<(*cell_iter).second->getCenter()[1]<<" "<<(*cell_iter).second->getCenter()[2]<<endl;

        if (obst->occupied_cells.count((*cell_iter).first) == 0)
        {
          Cell* newCell = (*cell_iter).second;
          vector< pair<long int,uint> >::iterator config_it;
          Vertex n;
          for (config_it =newCell->getConfigList()->begin(); config_it!=newCell->getConfigList()->end(); config_it++)
          {
            n = myPlanner->getID2VertexMap()->find((*config_it).first)->second;
            myPlanner->roadmap[n].config[(*config_it).second].invalid++;
            if (!(currentPath.empty()))
              for(path_iter=currentPath.begin(); path_iter!= currentPath.end(); path_iter++)
                if (n == (*path_iter))
                  myPlanner->notify();

          } //for
          obst->occupied_cells.insert(*cell_iter);
        }
      }//end of lists comparison

      cout<<"ObstaclesHandler::moveObstacle: Function duration "<<ros::Time::now() - start_time<<" sec"<<endl;
      return true;
    }
  //obstacle not found
  return false;
}

bool ObstaclesHandler::moveObstacle(Obstacle* obst, btVector3 newCenter)
{
  vector<Obstacle*>::iterator it;
  for (it=myObstacles.begin(); it!= myObstacles.end(); it++)
    if((*it) == obst)
    {
      return moveObstacle(obst,newCenter,obst->getTransform().getRotation());
    }
  return false;
}

bool ObstaclesHandler::moveObstacle(Obstacle* obst, btQuaternion newRot)
{
  vector<Obstacle*>::iterator it;
  for (it=myObstacles.begin(); it!= myObstacles.end(); it++)
    if((*it) == obst)
    {
      return moveObstacle(obst,obst->getTransform().getOrigin(), newRot);
    }
  return false;
}


bool ObstaclesHandler::deleteObstacle(Obstacle* obst)
{
  vector<Obstacle*>::iterator it;
  for (it=myObstacles.begin(); it!= myObstacles.end(); it++)
    if((*it) == obst)
    {
      map<long int, Cell*> cells = obst->occupied_cells;
      map<long int, Cell*>::iterator cell_iter = cells.begin();

      //validate all configs, that were occupied by the obstacle.
      for (cell_iter=obst->occupied_cells.begin(); cell_iter != obst->occupied_cells.end(); cell_iter++)
      {
       Cell* cellToValidate = (*cell_iter).second;
       vector< pair<long int,uint> >::iterator config_it;
       Vertex n;
       for (config_it =cellToValidate->getConfigList()->begin(); config_it!=cellToValidate->getConfigList()->end(); config_it++)
       {
         n = myPlanner->getID2VertexMap()->find((*config_it).first)->second;
         myPlanner->roadmap[n].config[(*config_it).second].invalid--;
       } //for
      }
      delete *it;
      myObstacles.erase(it);
      updateObstaclesRviz();
      return true;
    }
  return false;
}

long int ObstaclesHandler::FormIDfromPos(btVector3 position)
{
  //rounding down so we have number in format a.aaa
  int a,b,c;
  a = round(position[0]*1000);
  b = round(position[1]*1000);
  c = round(position[2]*1000);
  /*
   * ID_number structure: XxxxYyyyZzzzz
   * X - sign of centerPosition.x (1 for negative)
   * _xxx - value of centerPosition.x in mm (should be:  centerPosition.x <1)
   * Y - sign of centerPosition.y (1 for negative)
   * _yyy - value of centerPosition.y in mm. (should be:  centerPosition.x <1)
   * Z- sign of centerPosition.y (1 for negative)
   * _zzzz - value of centerPosition.z in mm
   */
  long int ID = abs(c) + abs(b*100000) + fabs(a*1e9);
  if (c < 0)
    ID += 1e4;
  if (b < 0)
    ID += 1e8;
  if (a < 0)
    ID += 1e12;
  return ID;
}

void ObstaclesHandler::updateObstaclesRviz(bool clear)
{
  visualization_msgs::MarkerArray obstacleMarkers;


  obstacleMarkers.markers.resize(myObstacles.size()-1);

  vector<Obstacle*>::iterator it=myObstacles.begin();
  btVector3 tmp_vec;
  btQuaternion tmp_trans;
  for (uint i=0; i<myObstacles.size()-1; i++)
  {
    if ((*it) == manipulator_base)
      it++;

    obstacleMarkers.markers[i].ns = "obstacles" ;
    obstacleMarkers.markers[i].header.stamp = ros::Time::now();
    obstacleMarkers.markers[i].header.frame_id = "/fix";
    obstacleMarkers.markers[i].id = i;
    if(clear)
      obstacleMarkers.markers[i].action = visualization_msgs::Marker::DELETE;
    else
      obstacleMarkers.markers[i].action = visualization_msgs::Marker::ADD;

    switch ((*it)->getType()){
      case (Obstacle::CUBE):
          tmp_vec = (static_cast<CubicObstacle*>(*it))->getDim();
          obstacleMarkers.markers[i].type =visualization_msgs::Marker::CUBE;
          break;
      case (Obstacle::SPHERE):
          tmp_vec = btVector3(1,1,1);
          tmp_vec *= 2*(static_cast<SphericObstacle*>(*it))->getRadius();
          obstacleMarkers.markers[i].type =visualization_msgs::Marker::SPHERE;
          break;
      default:
        obstacleMarkers.markers[i].action = visualization_msgs::Marker::DELETE;
        break;
    }
    obstacleMarkers.markers[i].scale.x = tmp_vec.getX();
    obstacleMarkers.markers[i].scale.y = tmp_vec.getY();
    obstacleMarkers.markers[i].scale.z = tmp_vec.getZ();

    tmp_vec = (*it)->getTransform().getOrigin();
    obstacleMarkers.markers[i].pose.position.x = tmp_vec.getX();
    obstacleMarkers.markers[i].pose.position.y = tmp_vec.getY();
    obstacleMarkers.markers[i].pose.position.z = tmp_vec.getZ();
    tmp_trans = (*it)->getTransform().getRotation();
    obstacleMarkers.markers[i].pose.orientation.x = tmp_trans.getX();
    obstacleMarkers.markers[i].pose.orientation.y = tmp_trans.getY();
    obstacleMarkers.markers[i].pose.orientation.z = tmp_trans.getZ();
    obstacleMarkers.markers[i].pose.orientation.w = tmp_trans.getW();

    obstacleMarkers.markers[i].lifetime = ros::Duration();
    obstacleMarkers.markers[i].color.r = 0.0f;
    obstacleMarkers.markers[i].color.g = .60f;
    obstacleMarkers.markers[i].color.b = .80f;
    obstacleMarkers.markers[i].color.a = 1.0;

    it++;
  }
  if (node_handler->ok())
    obstacle_publisher.publish(obstacleMarkers);
  if (node_handler->ok())
    obstacle_publisher.publish(obstacleMarkers);
}

vector<Obstacle*> ObstaclesHandler::getObstacles()
{
  return myObstacles;
}
