#include <ros/ros.h>
#include <ros/console.h>
#include <visualization_msgs/Marker.h>
#include "../matrixDefinition.h"
#include "../SchunkTransformhandler.h"
#include "../RobotVisualPublisher.h"
#include "../Cell.h"
#include "../Node.h"
#include "IKSolver.h"
#include "ConfToCellsMapper.h"
#include <fstream>
#include <boost/graph/adjacency_list.hpp>
#include <boost/graph/adj_list_serialize.hpp>
#include <boost/serialization/serialization.hpp>
#include <boost/serialization/vector.hpp>
//#include <boost/serialization/list.hpp>
#include <boost/serialization/map.hpp>




using namespace std;
using namespace boost;

long int FormIDfromPos(btVector3 position)
{
  //rounding down so we have number in format a.aaa
  int a,b,c;
  a = round(position[0]*1000);
  b = round(position[1]*1000);
  c = round(position[2]*1000);
  /*
   * ID_number structure: XxxxYyyyZzzzz
   * X - sign of centerPosition.x (1 for negative)
   * _xxx - value of centerPosition.x in mm (should be:  centerPosition.x <1)
   * Y - sign of centerPosition.y (1 for negative)
   * _yyy - value of centerPosition.y in mm. (should be:  centerPosition.x <1)
   * Z- sign of centerPosition.y (1 for negative)
   * _zzzz - value of centerPosition.z in mm
   */
  long int ID = abs(c) + abs(b*100000) + fabs(a*1e9);
  if (c < 0)
    ID += 1e4;
  if (b < 0)
    ID += 1e8;
  if (a < 0)
    ID += 1e12;
  return ID;
}


int main(int argc, char** argv) 
{
  ros::init(argc, argv, "offline_mapper");
  ros::NodeHandle n;
  std::cout<<std::setiosflags(std::ios::fixed) << std::setprecision(4);
  RobotVisualPublisher* rviz_publisher = new RobotVisualPublisher(&n);

  ros::Rate loop_rate(5);

  urdf::Model model;
  if (!model.initParam("robot_description")) {
    ROS_ERROR("Failed to parse urdf file");
    return -1;
  }
  ROS_INFO("Successfully parsed urdf file");

  cVector conf;
  btVector3 orient(0,1,0), point(0.2,-0.5,0.6);

  /*---------------MARKER------------*/
  ros::Publisher marker_pub = n.advertise<visualization_msgs::Marker>("visualization_marker", 1);

  visualization_msgs::Marker marker;
  marker.header.stamp = ros::Time::now();
  marker.header.frame_id = "/fix";
  marker.id = 0;
  marker.type =visualization_msgs::Marker::ARROW;
  marker.action = visualization_msgs::Marker::ADD;
  marker.pose.position.x = point.x();
  marker.pose.position.y = point.y();
  marker.pose.position.z = point.z();

  btQuaternion arrow_or;

  marker.lifetime = ros::Duration();
  marker.scale.x = 0.2;
  marker.scale.y = 0.5;
  marker.scale.z = 0.5;
  marker.color.r = 0.0f;
  marker.color.g = 1.0f;
  marker.color.b = 0.0f;
  marker.color.a = 1.0;

  /*---------------END MARKER------------*/

  /*--------------CELL ARRAY------------*/
  map<long int, Cell> cell_array;
  float min_x_cells = floor(MIN_X_RANGE/CELL_GRID_STEP)*CELL_GRID_STEP;
  float min_y_cells = floor(MIN_Y_RANGE/CELL_GRID_STEP)*CELL_GRID_STEP;
  float min_z_cells = floor(MIN_Z_RANGE/CELL_GRID_STEP)*CELL_GRID_STEP;
  float x=min_x_cells,y=min_y_cells,z=min_z_cells;
  ///double step = 0.05;
  while(z<=MAX_Z_RANGE)
  {
    while(y<=MAX_Y_RANGE)
    {
      while(x<=MAX_X_RANGE)
      {
        Cell buf_cell(btVector3(x,y,z));
        cell_array.insert(pair<long int, Cell>(buf_cell.getId(),buf_cell));
        x+=CELL_GRID_STEP;
      }
      y+=CELL_GRID_STEP;
      x=min_x_cells;
    }
    z+=CELL_GRID_STEP;
    y=min_y_cells;
  }
  cout<<"Cell array created"<<endl;
  /***********END CELL ARRAY***************/

  typedef adjacency_list<vecS,vecS,undirectedS,Node, Edge_prop, Graph_prop > RoadmapGraph;
  typedef graph_traits< RoadmapGraph >::vertex_descriptor RoadmapVertDesc;
  typedef graph_traits< RoadmapGraph >::vertex_iterator RoadmapVertIt;
  typedef graph_traits< RoadmapGraph >::edge_descriptor RoadmapEdgeDesc;
  typedef graph_traits< RoadmapGraph >::edge_descriptor RoadmapEdgeIt;

  ConfToCellsMapper mapper(&cell_array);
  IKSolver solver(model);

  RoadmapGraph roadmap;
  RoadmapVertDesc vertex;

  map<long int, RoadmapVertDesc> ID2VertexMap;

  float min_x_nodes = floor(MIN_X_RANGE/GRID_STEP)*GRID_STEP;
  float min_y_nodes = floor(MIN_Y_RANGE/GRID_STEP)*GRID_STEP;
  float min_z_nodes = floor(MIN_Z_RANGE/GRID_STEP)*GRID_STEP;
  x=min_x_nodes,y=min_y_nodes,z=min_z_nodes;
  //roadmap[graph_bundle].step_size = GRID_STEP;
  while(z<=1.2)
  {
    cout<<endl<<"z="<<z<<": ";
    while(y<=MAX_Y_RANGE)
    {
      cout<<".";
      while(x<=MAX_X_RANGE)
      {
        Node node;
        node.position = btVector3(x,y,z);
        node.FormIDfromPos();
        solver.computeIKForNode(node);
        mapper.mapNodeToCells(node,CELL_GRID_STEP);

        //if node is not empty, add it to the graph
        for (int i=0;i<6;i++)
          if(node.config[i].exist == true)
          {
            vertex = add_vertex(node,roadmap);
            ID2VertexMap.insert(pair<long int, RoadmapVertDesc>(node.ID,vertex));
            break;
          }


        x+=GRID_STEP;
      }

      y+=GRID_STEP;
      x=min_x_nodes;
    }

    z+=GRID_STEP;
    y=min_y_nodes;
  }
  cout<<endl;
  cout<<"Nodes created"<<endl;


  //****CREATING EDGES*****//
  RoadmapVertDesc vertex_from, vertex_to;
  RoadmapEdgeDesc edge;
  list<RoadmapVertDesc> vertices_to_remove;
  long int vertexID;
  x=min_x_nodes,y=min_y_nodes,z=min_z_nodes;
  while(z<=MAX_Z_RANGE)
  {
    while(y<=MAX_Y_RANGE)
    {
      while(x<=MAX_X_RANGE)
      {
        vertexID = FormIDfromPos(btVector3(x,y,z));
        if (ID2VertexMap.count(vertexID) != 0)
          vertex_from = ID2VertexMap.find(vertexID)->second;
        else
        {
          int yint = (int)(vertexID/1e5) - ((int)(vertexID/1e8))*10000;
          int xint = (int)(vertexID/1e9) - ((int)(vertexID/1e12))*1000;
          int zint = vertexID - ((int)(vertexID/10000))*10000;
          int div = (int) (GRID_STEP*1000);
          if (zint%div !=0 || xint%div !=0 || yint%div!=0)
            cout<<"node "<<vertexID<<" (pos: "<<x<<" "<<y<<" "<<z<<") not found"<< endl;
          x+=GRID_STEP;
          continue;
        }


        if (x+GRID_STEP<=0.8)
        {
          vertexID = FormIDfromPos(btVector3(x+GRID_STEP,y,z));
          if (ID2VertexMap.count(vertexID) != 0)
          {
            vertex_to = ID2VertexMap.find(vertexID)->second;
            add_edge(vertex_from,vertex_to,Edge_prop(1),roadmap);
          }
        }
        if (y+GRID_STEP<=0.8)
        {
          vertexID = FormIDfromPos(btVector3(x,y+GRID_STEP,z));
          if (ID2VertexMap.count(vertexID) != 0)
          {
            vertex_to = ID2VertexMap.find(vertexID)->second;
            add_edge(vertex_from,vertex_to,Edge_prop(1),roadmap);
          }
        }
        if (z+GRID_STEP<=1.1)
        {
          vertexID = FormIDfromPos(btVector3(x,y,z+GRID_STEP));
          if (ID2VertexMap.count(vertexID) != 0)
          {
            vertex_to = ID2VertexMap.find(vertexID)->second;
            add_edge(vertex_from,vertex_to,Edge_prop(1),roadmap);
          }
        }

        x+=GRID_STEP;
      }
      y+=GRID_STEP;
      x=min_x_nodes;
    }
    z+=GRID_STEP;
    y=min_y_nodes;
  }

  cout<<"Edges inside the graph created"<<endl;

//****SAVING THE GRAPH INTO A FILE****///
  if(true)
  {
    ofstream myfile;
    myfile.open ("/home/sergey/kunz_brock_planner/data/Roadmap.dat");
    archive::text_oarchive oa(myfile);
    oa<<roadmap;
    myfile.close();
    cout<<"Graph export completed"<<endl;

    myfile.open ("/home/sergey/kunz_brock_planner/data/ID2VertexMap.dat");
    oa<<ID2VertexMap;
    myfile.close();
    cout<<"ID2VertexMap export completed"<<endl;

    myfile.open ("/home/sergey/kunz_brock_planner/data/Cells.dat");
    oa<<cell_array;
    myfile.close();
    cout<<"Cell array export completed"<<endl;

  }

//  /****PRINTING GRAPH STRUCTURE*****/
//  graph_traits < RoadmapGraph >::adjacency_iterator ai, a_end;
//  cout<<"vertices:"<<endl;
//  RoadmapVertIt iter, end;
//  for (tie(iter, end) = vertices(roadmap); iter != end; ++iter)
//  {
//    cout<<"Node: "<<*iter<<" ID: "<<roadmap[*iter].ID<<endl;
//    //cout<<roadmap[0]
//    tie(ai, a_end) = adjacent_vertices(*iter,roadmap);
//    if (ai == a_end)
//      cout << " has no edges";
//    else
//      cout << " has edges with ";
//    for (; ai != a_end; ++ai) {
//     cout << *ai;
//      if (next(ai) != a_end)
//        cout << ", ";
//    }
//    cout<<endl;
//  }

  /***access to edges and their properties***/
//  RoadmapGraph::out_edge_iterator edge_it,edge_end;
//  pair<RoadmapGraph::out_edge_iterator,RoadmapGraph::out_edge_iterator> a;
//  a = out_edges(0,roadmap);
//  edge_it = a.first;
//  RoadmapGraph::edge_descriptor b = *(out_edges(0,roadmap).first);
//  cout<<roadmap[*edge_it].weight;
//  cout<<roadmap[b].weight;



//  map<long int, Node> nodes;
//
//
//  while (point.m_floats[1]<=0.5)
//  {
//    point.m_floats[1]+=0.1;
//    Node node;
//    node.position = point;
//    node.FormIDfromPos();
//    solver.computeIKForNode(node);
//    mapper.mapNodeToCells(node,GRID_STEP);
//    nodes.insert(pair<long int, Node>(node.ID, node));
//  }
//  while (ros::ok()) {


//    marker.type =visualization_msgs::Marker::ARROW;
//    marker.scale.x = 0.2;
//    marker.scale.y = 0.5;
//    marker.scale.z = 0.5;
//    marker.pose.position.x = point.x();
//    marker.pose.position.y = point.y();
//    marker.pose.position.z = point.z();
//    bool ik_success = solver.solveIKtask(point,orient,conf);
//    if (orient[0]==0)
//      arrow_or = btQuaternion(btVector3(0.,0.,1.),M_PI_2*orient[1])*btQuaternion(btVector3(0.,1.,0.),-M_PI_2*orient[2]);
//    else
//      arrow_or = btQuaternion(btVector3(0.,1.,0.),M_PI_2*(1 - orient[0]));
//    marker.pose.orientation.x = arrow_or.getX();
//    marker.pose.orientation.y = arrow_or.getY();
//    marker.pose.orientation.z = arrow_or.getZ();
//    marker.pose.orientation.w = arrow_or.getW();
//
//    if (ik_success)
//    {
//      rviz_publisher->PublishJointState(conf);
//      cout<<"Point "<<point[0]<<" "<<point[1]<<" "<<point[2]<<" "<<" solved."<<endl<<conf.transpose()<<endl;
//
//      Configuration a(conf);
//      //mapper.mapConfiguratuonToCells(a,GRID_STEP);
//      set<long int> conf_list = a.getCells();
//      set<long int>::iterator iter;
//
//      marker.type =visualization_msgs::Marker::CUBE;
//       marker.scale.x = GRID_STEP;
//       marker.scale.y = GRID_STEP;
//       marker.scale.z = GRID_STEP;
//
//      for(iter = conf_list.begin();iter!=conf_list.end();iter++)
//      {
//        btVector3 cent = (cell_array.find(*iter))->second.getCenter();
//        marker.pose.position.x = cent.x();
//        marker.pose.position.y = cent.y();
//        marker.pose.position.z = cent.z();
//        marker_pub.publish(marker);
//        usleep(200000);
//      }
//
//    }
//    else
//      cout<<"Point "<<point[0]<<" "<<point[1]<<" "<<point[2]<<" not solved."<<endl;
//    marker_pub.publish(marker);




//    pair<long int,Cell> a = *cell_array.find(30010050);
//    cout<<cell_array.count(100)<<endl;
//    vector< pair<long int,uint> > configs = a.second.getConfigList();
//    cout<<configs.size();
//    vector< pair<long int,uint> >::iterator vect_it;
//    for(vect_it = configs.begin(); vect_it!=configs.end(); vect_it++)
//    {
//      if (ID2VertexMap.count((*vect_it).first) != 0)
//      {
//        RoadmapVertDesc sdf = (*ID2VertexMap.find((*vect_it).first)).second;
//        Node k = roadmap[sdf];
//        uint l=(*vect_it).second;
//        conf = k.config[l].getCvector();
//        rviz_publisher->PublishJointState(conf);
//        cout<<"Node "<<k.ID<<", config "<<l<<endl;
//        loop_rate.sleep();
//      }
//      else
//        cout<<"node "<< (*ID2VertexMap.find((*vect_it).first)).first <<" not found"<<endl;
//    }

  cout<<"Offline mapper execution finished normally"<<endl;
  delete rviz_publisher;

  return 0;
}
