/*
 * Planner.cpp
 *
 *  Created on: 16.04.2012
 *      Author: sergey
 */

#include "Planner.h"
#include "Cell.h"
#include "Node.h"
#include <fstream>
#include <boost/graph/adjacency_list.hpp>
#include <boost/graph/adj_list_serialize.hpp>
#include <boost/serialization/serialization.hpp>
#include <boost/serialization/vector.hpp>
#include <boost/serialization/map.hpp>
#include <boost/graph/astar_search.hpp>
#include <list>
#include <boost/date_time/posix_time/posix_time.hpp>
#include <pthread.h>
#include <visualization_msgs/MarkerArray.h>
#include "Behaviours/EndEffectorBehavior.h"
#include "Behaviours/AvoidanceBehaviour.h"
#include "Behaviours/GlobalBehaviour.h"

using namespace std;
using namespace boost;
using namespace boost::posix_time;

RoadmapGraph Planner::roadmap;

Planner::Planner(ros::NodeHandle* n)
{


  ReadyForExitFlag = false;
  /***LOADING PRECOMPUTED DATA***/
  ifstream myfile;
  myfile.open ("/home/sergey/kunz_brock_planner/data/Roadmap.dat");
  archive::text_iarchive ia(myfile);
  ia>>roadmap;
  myfile.close();
  cout<<"Roadmap graph import completed"<<endl;
  cout<<"Loading cell array. Please, be patient."<<endl;

  myfile.open ("/home/sergey/kunz_brock_planner/data/Cells.dat");
  ia>>cell_array;
  myfile.close();
  cout<<"Cell array import completed"<<endl;

  myObstacles = new ObstaclesHandler(n, this);

  myfile.open ("/home/sergey/kunz_brock_planner/data/ID2VertexMap.dat");
  ia>>ID2VertexMap;
  myfile.close();
  cout<<"ID2VertexMap import completed"<<endl;

  performTask = false;
  findNewPath = true;
  replanningCounter = 0;
  myRobot = new RobotController(n);
  myRobot->setObstacleHandler(myObstacles);

  markerarray_pub = n->advertise<visualization_msgs::MarkerArray>("visualization_marker_array", 1);
  pthread_create(new pthread_t, 0, _run, this);
}

Planner::~Planner()
{
  delete myRobot;
}

void Planner::setNewGlobalGoal(btVector3 goal)
{
  goalPos = goal;
  findNewPath = true;
  performTask = true;
}

void Planner::setNewEEaimingGoal(btVector3 goal)
{
  myRobot->getEndEffectorBehavior()->setGoalPoint(goal);
}

void Planner::setEEOrientationGoal(btVector3 orient)
{
  myRobot->getEndEffectorBehavior()->setGoalOrientation(orient);
}

bool Planner::checkExecutionStatus()
{
  return ReadyForExitFlag;
}

/**
   @brief Static "dummy function" for pthread_create
*/
void* Planner::_run(void* arg) {
  reinterpret_cast<Planner*>(arg)->run();
  return 0;
}

/**
   @brief Main loop. Handles the main steps of the motion planning:
   Plotting route to goal and making sure route is followed
*/
void Planner::run()
{
  ptime now, then;
  btVector3 currentPos;

  while(true) {
    then = microsec_clock::local_time();

    if (performTask)
    {
      vector<btVector3> origins = myRobot->getTransformhandler()->getOrigins();
      currentPos = origins[origins.size()-1];
      // Got an initial and goal milestone
      if (findNewPath)
      {
        if (find_path(currentPos, goalPos, currentPath))
        {
          bestDistanceToGoal = HUGE_VALF;
          tryCounter = 0;
          findNewPath = false;
          std::cout << endl<<"A valid path found"<<endl;
          myRobot->setPositionGoal(*(currentPath.begin()));

          //****print the path****///
          cout << "Shortest path: start: "<<currentPos.getX()<<" "<<currentPos.getY()<<" "<<currentPos.getZ()<<" "<<endl;
          list<btVector3>::iterator spi = currentPath.begin();
          cout << " path: "<<endl;
          for(++spi; spi != currentPath.end(); ++spi)
            cout << "-> (" << spi->getX()<<"; "<< spi->getY()<<"; "<<spi->getZ()<<") "<<endl;
          cout<<endl;
          btVector3 nextWaypoint1 = *(++currentPath.begin());
          cout<<"nextWaypoint: "<<nextWaypoint1.getX()<<" "<<nextWaypoint1.getY()<<" "<<nextWaypoint1.getZ()<<" "<<endl;

          VisualizePath();
          VisualizePath();
          replanningCounter = 0;
        }
        else
        {
          replanningCounter++;
          std::cout << "Planner::run(): WARNING: Could not find a valid path. Trying again..\n"<<endl;
          findNewPath = true;
        }
        //if we can't find path for too long, refresh all invalidated nodes and try again
        if (replanningCounter >= 10)
        {
          NodesBlackList.clear();

          replanningCounter = 0;
        }
      }
      else
      {
        // Got a path, now make sure to follow it
        btVector3 nextWaypoint = *(currentPath.begin());
        lastEdgeInPath = (currentPath.size() == 1);

        // If we're close enough to the target config, move along the next edge in the path
        // else if its the last edge in path and we're close enough: clear myTarget

        // Finding distance to goal
        double distanceToGoal = (currentPos - nextWaypoint).length();

        tryCounter++;
        ///failure detection: if a distance to milestone isn't decreasing, delete the edge
        if (distanceToGoal < bestDistanceToGoal)
        {
                bestDistanceToGoal = distanceToGoal;
                tryCounter = 0;
        }
        if (tryCounter >= 7)
        {
          cout<<"No progress towards goal detected"<<endl;

          //put the weight of edges, incoming to the node, as a high value
          Vertex k;
          MapPosToClosestNode(nextWaypoint,k);
          NodesBlackList.push_back(k);
          cout<<"NODE "<<k<<"("<<roadmap[k].position[0]<<" "<<roadmap[k].position[1]
               <<" "<<roadmap[k].position[2]<< ") ADDED TO BLACKLIST"<<endl;

          myRobot->getGlobalBehaviour()->disableBehaviour();
          findNewPath = true;
          currentPath.clear();
          ClearPathRVIZ();
          continue;
        }
        //std::cout << "Distance to goal = " << distanceToGoal << std::endl;

        if (lastEdgeInPath && distanceToGoal < WAYPOINT_ACCURACY/2)
        {
          //We're done
          currentPath.clear();
          currentPathNodes.clear();
          performTask = false;
          cout<<"PATH COMPLETED"<<endl;

          NodesBlackList.clear();

          ReadyForExitFlag  = true;
          break;
        }
        else
          if (!lastEdgeInPath && distanceToGoal < WAYPOINT_ACCURACY*2)
          {
            // Move on towards next milestone in our path (deleting the first edge in the path list)
            currentPath.pop_front();
            if (!currentPathNodes.empty())
              currentPathNodes.pop_front();
            myRobot->setPositionGoal(*(currentPath.begin()));
            bestDistanceToGoal = HUGE_VALF;
            tryCounter = 0;
          }
      }

    } else {
      // No goal: nothing to do
    }
  //  myObstacles->updateObstaclesRviz();
    now = microsec_clock::local_time();
    usleep((microseconds(1000000.0/PLANNER_FREQ) - (now - then)).total_microseconds());
  }

  pthread_exit(0);
}

/**
   @brief Notify of changes in the world that could affect the planned
   motion between milestones.
   @example A new milestone or obstacle is added or an edge has changed
   its status.
*/
void Planner::notify(void) {
  findNewPath = true;
}

/**
 * @brief checks if the node in the graph can be used in the path.
 * returns true, if valid, false otherwise.
 */

bool Planner::ExamineNode (Vertex u)
{
  uint EETaskType;
  btVector3 EETaskVector = myRobot->getEndEffectorBehavior()->getGoal(EETaskType);

  btVector3 desiredEEaxis;

  if (pow(roadmap[u].position.getX(),2) + pow(roadmap[u].position.getY(),2) < pow(0.07,2)
      && roadmap[u].position.getZ() < 0.3)
    return false;

  list<Vertex>::iterator vertex_it;
  for (vertex_it=NodesBlackList.begin(); vertex_it!=NodesBlackList.end(); vertex_it++)
    if (*vertex_it == u)
    {
      cout<<"vertex "<<u<<" is in the blacklist"<<endl;
      return false;
    }
  switch (EETaskType){
    case (EndEffectorBehavior::GOAL_PRESERVE_ORIENTATION):
        desiredEEaxis = EETaskVector;
        break;
    case (EndEffectorBehavior::GOAL_AIM_AT_POINT):
        desiredEEaxis= (EETaskVector - roadmap[u].position).normalize();
        break;
    default:
      //without end effector task any node that has valid configuration can be used
      return ((roadmap[u].config[Node::X_POS].exist && roadmap[u].config[Node::X_POS].invalid == 0)) ||
          (roadmap[u].config[Node::X_NEG].exist && roadmap[u].config[Node::X_NEG].invalid == 0) ||
          (roadmap[u].config[Node::Y_POS].exist && roadmap[u].config[Node::Y_POS].invalid == 0) ||
          (roadmap[u].config[Node::Y_NEG].exist && roadmap[u].config[Node::Y_NEG].invalid == 0) ||
          (roadmap[u].config[Node::Z_POS].exist && roadmap[u].config[Node::Z_POS].invalid == 0) ||
          (roadmap[u].config[Node::Z_NEG].exist && roadmap[u].config[Node::Z_NEG].invalid == 0);
  }


  bool valid = true;

 // Node k = roadmap[u];
  double treshold = 0.56;

  //"configuration,invalid == 0" means, that configuration is obstacle-free
  if (desiredEEaxis.getX() > treshold)
    valid = valid && (roadmap[u].config[Node::X_POS].invalid == 0) && roadmap[u].config[Node::X_POS].exist;
  if (desiredEEaxis.getY() > treshold)
    valid = valid && (roadmap[u].config[Node::Y_POS].invalid == 0) && roadmap[u].config[Node::Y_POS].exist;
  if (desiredEEaxis.getZ() > treshold)
    valid = valid && (roadmap[u].config[Node::Z_POS].invalid == 0) && roadmap[u].config[Node::Z_POS].exist;
  if (desiredEEaxis.getX() < -treshold)
    valid = valid && (roadmap[u].config[Node::X_NEG].invalid == 0) && roadmap[u].config[Node::X_NEG].exist;
  if (desiredEEaxis.getY() < -treshold)
    valid = valid && (roadmap[u].config[Node::Y_NEG].invalid == 0) && roadmap[u].config[Node::Y_NEG].exist;
  if (desiredEEaxis.getZ() < -treshold)
    valid = valid && (roadmap[u].config[Node::Z_NEG].invalid == 0) && roadmap[u].config[Node::Z_NEG].exist;

  //cout<<"vertex "<<u<<": "<<roadmap[u].config[Node::X_POS].invalid<<"; ";

  return valid;
}

bool Planner::find_path (btVector3 cur_pos, btVector3 goal_pos, list<btVector3>& result)
{
  //cout<<"current pos: "<<cur_pos[0]<<" "<<cur_pos[1]<<" "<<cur_pos[2]<<endl;
  list<btVector3> shortest_path;
  Vertex start, goal;
  currentPathNodes.clear();
  MapPosToClosestNode(cur_pos, start);

  list<Vertex>::iterator blacklist_it;
  for (blacklist_it = NodesBlackList.begin();blacklist_it!=NodesBlackList.end();blacklist_it++)
    if (start == *blacklist_it)
    {
      MapPosToClosestNode(cur_pos+btVector3(GRID_STEP,0,0), start);
      for (blacklist_it = NodesBlackList.begin();blacklist_it!=NodesBlackList.end();blacklist_it++)
        if (start == *blacklist_it)
        {
          MapPosToClosestNode(cur_pos+btVector3(-GRID_STEP,0,0), start);
          for (blacklist_it = NodesBlackList.begin();blacklist_it!=NodesBlackList.end();blacklist_it++)
            if (start == *blacklist_it)
            {
              MapPosToClosestNode(cur_pos+btVector3(0,GRID_STEP,0), start);
              for (blacklist_it = NodesBlackList.begin();blacklist_it!=NodesBlackList.end();blacklist_it++)
                if (start == *blacklist_it)
                {
                  MapPosToClosestNode(cur_pos+btVector3(0,-GRID_STEP,0), start);
                  for (blacklist_it = NodesBlackList.begin();blacklist_it!=NodesBlackList.end();blacklist_it++)
                    if (start == *blacklist_it)
                    {
                      MapPosToClosestNode(cur_pos+btVector3(0,0,GRID_STEP), start);
                      for (blacklist_it = NodesBlackList.begin();blacklist_it!=NodesBlackList.end();blacklist_it++)
                        if (start == *blacklist_it)
                        {
                          MapPosToClosestNode(cur_pos+btVector3(0,0,-GRID_STEP), start);
                          for (blacklist_it = NodesBlackList.begin();blacklist_it!=NodesBlackList.end();blacklist_it++)
                            if (start == *blacklist_it)
                            {
                              start=0;
                              break;
                            }

                          break;
                        }

                      break;
                    }

                  break;
                }

              break;
            }

          break;
        }
      break;
    }

  if (!(MapPosToClosestNode(cur_pos, start)) || !(MapPosToClosestNode(goal_pos, goal)))
  {
    cout<<"Start or goal position are out of reach"<<endl;
    shortest_path.clear();
    result = shortest_path;
    return false;
  }
  vector<Vertex> p(num_vertices(roadmap));

  WeightMap weights = get(&Edge_prop::weight, roadmap);
  vector<double> cost(num_vertices(roadmap));
  ros::Time start_time = ros::Time::now();

  try {
      // call astar named parameter interface
      astar_search
        (roadmap, start,
         distance_heuristic(goal,roadmap),
         predecessor_map(&p[0]).
         weight_map(weights).
         distance_map(&cost[0]).
         visitor(astar_goal_visitor(goal, weights,invalidatedEdgesForPlanning, this)));
    }

    // found a path to the goal
    catch(found_goal fg)
    {
      //copy the path
      for(Vertex v = goal;; v = p[v])
      {
        shortest_path.push_front(roadmap[v].position);
        currentPathNodes.push_front(v);
        if(p[v] == v)
          break;
      }
      //shortest_path.push_front(cur_pos);
      shortest_path.push_back(goal_pos);
      shortest_path.push_back(goal_pos);
      PathLength = shortest_path.size();
      cout<<"Planner::find_path: Search duration "<<ros::Time::now() - start_time<<" sec"<<endl;
      //remove the changes made during the search
      list<RoadmapEdgeDesc>::iterator ie_it = invalidatedEdgesForPlanning.end();
      for(--ie_it; ie_it != invalidatedEdgesForPlanning.begin(); --ie_it)
      {
          put(weights,*ie_it,1);
      }
      invalidatedEdgesForPlanning.clear();
  }

  if (cost[goal]>= 20000)
  {
    cout<< "No valid path found"<<endl;

    cout<<"Planner::find_path: Search duration "<<ros::Time::now() - start_time<<" sec"<<endl;
    shortest_path.clear();
    result = shortest_path;
    return false;
  }
  else
  {
//    //print the path
//    cout << "Shortest path from " << start << " to "
//         << goal << ": ";
//    list<btVector3>::iterator spi = shortest_path.begin();
//    cout << "start"<<endl;
//    for(++spi; spi != shortest_path.end(); ++spi)
//      cout << "-> (" << spi->getX()<<"; "<< spi->getY()<<"; "<<spi->getZ()<<") "<<endl;
//    cout<<endl;

    result = shortest_path;
    return true;
  }
}


bool Planner::MapPosToClosestNode(btVector3 position, Vertex& node)
{
  //cout<<"Original pos: "<< position[0]<<" "<< position[1]<<" "<< position[2]<<endl;
  if ((fabs(position[0])>HORIZONTAL_DISTANCE_LIMIT) ||
      (fabs(position[1])>HORIZONTAL_DISTANCE_LIMIT) ||
      (position[2] > VERTICAL_LIMIT_MAX) ||
      (position[2] < VERTICAL_LIMIT_MIN))
    return false;

  btVector3 buf;
  buf[0] = int((position[0]+GRID_STEP/2)/GRID_STEP)*GRID_STEP;
  buf[1] = int((position[1]+GRID_STEP/2)/GRID_STEP)*GRID_STEP;
  buf[2] = int((position[2]+GRID_STEP/2)/GRID_STEP)*GRID_STEP;
//if (buf[0] - position[0])

  //rounding down
  int a = round(buf[0]*1000);
  int b = round(buf[1]*1000);
  int c = round(buf[2]*1000);
  /*
   * ID_number structure: XxxxYyyyZzzzz
   * X - sign of centerPosition.x (1 for negative)
   * _xxx - value of centerPosition.x in mm (should be:  centerPosition.x <1)
   * Y - sign of centerPosition.y (1 for negative)
   * _yyy - value of centerPosition.y in mm. (should be:  centerPosition.x <1)
   * Z- sign of centerPosition.y (1 for negative)
   * _zzzz - value of centerPosition.z in mm
   */
  long int ID = abs(c) + abs(b*100000) + fabs(a*1e9);
  if (c < 0)
    ID += 1e4;
  if (b < 0)
    ID += 1e8;
  if (a < 0)
    ID += 1e12;

  if (ID2VertexMap.count(ID) != 0)
  {
    node = ID2VertexMap.find(ID)->second;
    return true;
  }
  else
  {
    cout<<"WARNING: Planner::MapPosToClosestNode: Can't find node with ID="<<ID<<endl;
    return false;
  }

  //cout<<"Node pos: "<< roadmap[ret].position[0]<<" "<< roadmap[ret].position[1]<<" "<< roadmap[ret].position[2]<<endl;

}

void Planner::VisualizePath(bool clear)
{

  visualization_msgs::MarkerArray pathMarkers;
  pathMarkers.markers.resize(currentPath.size());

  list<btVector3>::iterator it=currentPath.begin();

  btQuaternion tmp_trans;

  for (uint i=0; i<currentPath.size(); i++)
  {
    pathMarkers.markers[i].ns = "Path" ;
    pathMarkers.markers[i].header.stamp = ros::Time::now();
    pathMarkers.markers[i].header.frame_id = "/fix";
    pathMarkers.markers[i].id = i;
    pathMarkers.markers[i].type =visualization_msgs::Marker::SPHERE;
    if (clear == false)
      pathMarkers.markers[i].action = visualization_msgs::Marker::ADD;
    else
      pathMarkers.markers[i].action = visualization_msgs::Marker::DELETE;
    pathMarkers.markers[i].pose.position.x = (*it).getX();
    pathMarkers.markers[i].pose.position.y = (*it).getY();
    pathMarkers.markers[i].pose.position.z = (*it).getZ();
    tmp_trans = btQuaternion(btVector3(0,0,1),0);
    pathMarkers.markers[i].pose.orientation.x = tmp_trans.getX();
    pathMarkers.markers[i].pose.orientation.y = tmp_trans.getY();
    pathMarkers.markers[i].pose.orientation.z = tmp_trans.getZ();
    pathMarkers.markers[i].pose.orientation.w = tmp_trans.getW();

    pathMarkers.markers[i].scale.x = 0.02;
    pathMarkers.markers[i].scale.y = 0.02;
    pathMarkers.markers[i].scale.z = 0.02;

    pathMarkers.markers[i].lifetime = ros::Duration();
    pathMarkers.markers[i].color.r = 0.0f;
    pathMarkers.markers[i].color.g = 1.0f;
    pathMarkers.markers[i].color.b = 0.0f;
    pathMarkers.markers[i].color.a = 1.0;

    it++;
  }
  /*---------------END MARKER------------*/
  if (ros::ok())
    markerarray_pub.publish(pathMarkers);
}

void Planner::ClearPathRVIZ()
{
  visualization_msgs::MarkerArray pathMarkers;
  pathMarkers.markers.resize(currentPath.size());

  list<btVector3>::iterator it=currentPath.begin();

  for (uint i=0; i<currentPath.size(); i++)
    {
      pathMarkers.markers[i].ns = "Path" ;
      pathMarkers.markers[i].header.stamp = ros::Time::now();
      pathMarkers.markers[i].header.frame_id = "/fix";
      pathMarkers.markers[i].id = i;
      pathMarkers.markers[i].type =visualization_msgs::Marker::SPHERE;

      pathMarkers.markers[i].action = visualization_msgs::Marker::DELETE;

      it++;
    }
  if (ros::ok())
    markerarray_pub.publish(pathMarkers);
}



ObstaclesHandler* Planner::getObstaclesHandler()
{
  return myObstacles;
}

map<long int, Cell>* Planner::getCellArray()
{
  return &cell_array;
}

map<long int, Vertex>* Planner::getID2VertexMap()
{
  return &ID2VertexMap;
}

RobotController* Planner::getRobotController()
{
  return myRobot;
}

list<btVector3> Planner::getCurrentPath()
{
  list<btVector3> ret;
  pthread_mutex_lock(&pathMutex);
  ret = currentPath;
  pthread_mutex_unlock(&pathMutex);
  return ret;
}


list<Vertex> Planner::getCurrentPathNodes()
{
  list<Vertex> ret;
  pthread_mutex_lock(&pathMutex);
  ret = currentPathNodes;
  pthread_mutex_unlock(&pathMutex);
  return ret;
}











