/*
 * Planner.h
 *
 *  Created on: 16.04.2012
 *      Author: sergey
 */

#ifndef PLANNER_H_
#define PLANNER_H_

#include "Cell.h"
#include "Node.h"
#include "RobotController.h"
#include "ObstaclesHandler.h"
#include <boost/graph/adjacency_list.hpp>
#include <map>
#include <boost/graph/astar_search.hpp>

//#define GRID_STEP 0.05
#define HORIZONTAL_DISTANCE_LIMIT 0.8
#define VERTICAL_LIMIT_MAX 1.15
#define VERTICAL_LIMIT_MIN -0.15
#define WAYPOINT_ACCURACY 0.02

using namespace std;
using namespace boost;

typedef adjacency_list<vecS,vecS,undirectedS,Node, Edge_prop, Graph_prop > RoadmapGraph;
typedef graph_traits< RoadmapGraph >::vertex_descriptor Vertex;
typedef graph_traits< RoadmapGraph >::vertex_iterator RoadmapVertIt;
typedef graph_traits< RoadmapGraph >::edge_descriptor RoadmapEdgeDesc;
typedef graph_traits< RoadmapGraph >::edge_descriptor RoadmapEdgeIt;
typedef property_map<RoadmapGraph, double Edge_prop::*>::type WeightMap;

#define PLANNER_FREQ 5.0

class ObstaclesHandler;

class Planner
{
public:
  Planner(ros::NodeHandle* n);
  virtual ~Planner();
  static RoadmapGraph roadmap;
  bool ExamineNode (Vertex u);
  bool find_path (btVector3 cur_pos, btVector3 goal_pos, list<btVector3>& result);
  bool MapPosToClosestNode(btVector3 position, Vertex& node);
  void notify(void);
  void setNewGlobalGoal(btVector3 goal);
  void setNewEEaimingGoal(btVector3 goal);
  void setEEOrientationGoal (btVector3 orient);

  bool checkExecutionStatus();
  RobotController* getRobotController();
  ObstaclesHandler* getObstaclesHandler();
  map<long int, Cell>* getCellArray();
  map<long int, Vertex>* getID2VertexMap();
  list<btVector3> getCurrentPath();
  list<Vertex> getCurrentPathNodes();
private:
  static void* _run(void*);
  void run();
  void VisualizePath(bool clear = false);
  void ClearPathRVIZ();

  bool ReadyForExitFlag;

  //data to detect a "no progress toward a goal" failure
  double bestDistanceToGoal;
  int tryCounter;
  int replanningCounter;

  pthread_mutex_t pathMutex;

  list<btVector3> currentPath;
  list<Vertex> currentPathNodes;
  btVector3 goalPos;
  uint PathLength;
  bool findNewPath, performTask, lastEdgeInPath;

  map<long int, Cell> cell_array;
  list<RoadmapEdgeDesc> invalidatedEdgesForPlanning;
  list<Vertex> NodesBlackList;
  map<long int, Vertex> ID2VertexMap;
  RobotController* myRobot;
  ObstaclesHandler* myObstacles;
  ros::Publisher markerarray_pub;

  //------------------------------------------------------------------------///
  ///************A*-search, Boost Graph Library auxiliary functions*********///
  //------------------------------------------------------------------------///
  // euclidean distance heuristic
  class distance_heuristic : public astar_heuristic<RoadmapGraph, double>
  {
  public:
    distance_heuristic(Vertex goal,RoadmapGraph& g)
    {goal_pos = roadmap[goal].position;}
    double operator()(Vertex u)
    {
      btVector3 pos = roadmap[u].position;
      return (pos - goal_pos).length();
    }
  private:
    btVector3 goal_pos;
  };

  //exception when the path is found
  struct found_goal {};

  // Visitor function (calls check for validity and task-consistency; drives the search away from invalid nodes)
  class astar_goal_visitor : public default_astar_visitor
  {
  public:
    astar_goal_visitor(Vertex goal, WeightMap& w, list<RoadmapEdgeDesc>& ie, Planner* planner)
    : m_goal(goal), weights(w), invalid_edges(ie), owningPlanner(planner) {    }

    void discover_vertex(Vertex u, const RoadmapGraph& g) {
      if(u == m_goal)
        throw found_goal();
    }

    void examine_vertex(Vertex u, const RoadmapGraph& g) {
        illegal = !owningPlanner->ExamineNode(u);
    }

    void examine_edge(RoadmapEdgeDesc e, const RoadmapGraph& g) {
      if (illegal)
      {
        put(weights,e,  60000);
        invalid_edges.push_back(e);
      }
    }

  private:
    Vertex m_goal;
    bool illegal;
    WeightMap& weights;
    list<RoadmapEdgeDesc>& invalid_edges;
    Planner* owningPlanner;
  };



};

#endif /* PLANNER_H_ */
