/*
 * RobotController.cpp
 *
 *  Created on: 19.04.2012
 *      Author: sergey
 */

#include "RobotController.h"
#include "SchunkTransformhandler.h"
#include "TaskLevelController.h"
#include "Behaviours/GlobalBehaviour.h"
#include "Behaviours/StopBehavior.h"
#include "Behaviours/KinematicBehaviour.h"
#include "Behaviours/EndEffectorBehavior.h"
#include "Behaviours/AvoidanceBehaviour.h"
#include "Behaviours/SecondaryBehavior.h"
#include "kunz_brock_planner/EEpos.h"

RobotController::RobotController(ros::NodeHandle* n)
{
  myTransformhandler = new SchunkTransformhandler();
  myTLC = new TaskLevelController(this,7);
  rvizPublisher = new RobotVisualPublisher(n);
  EndEffectorPosPublisher = n->advertise<kunz_brock_planner::EEpos>("EE_position", 2);

  node_handler = n;

  myGlobalBehaviour = new GlobalBehaviour(myTransformhandler,myTLC);
  myEndEffectorBehavior = new EndEffectorBehavior(myTransformhandler,myTLC);
  myAvoidanceBehaviours.push_back(new AvoidanceBehaviour(myTransformhandler,myTLC,3,0.03,n));
  myAvoidanceBehaviours.push_back(new AvoidanceBehaviour(myTransformhandler,myTLC,5,0.03,n));
  myAvoidanceBehaviours.push_back(new AvoidanceBehaviour(myTransformhandler,myTLC,6,0.03,n));

  myTLC->addBehaviour(new KinematicBehaviour(myTransformhandler,myTLC));
  for (uint i=0; i<myAvoidanceBehaviours.size();i++)
    myTLC->addBehaviour(myAvoidanceBehaviours[i]);
  myTLC->addBehaviour(myEndEffectorBehavior);
  myTLC->addBehaviour(myGlobalBehaviour);
  myTLC->addBehaviour(new SecondaryBehavior(myTransformhandler,myTLC));
  myTLC->addBehaviour(new StopBehavior(myTransformhandler,myTLC));


  cVector initConf;
  initConf << 0, M_PI/4, 0, M_PI/4, 0, -M_PI/4, 0;
  myTransformhandler->setConfiguration(initConf);

  pthread_create(new pthread_t, 0, _run, this);
}

RobotController::~RobotController()
{
  delete myTransformhandler;
  delete myTLC;
  delete rvizPublisher;
}

void RobotController::setPositionGoal(btVector3 goal)
{
  myGlobalBehaviour->setGoalPoint(goal);
}

GlobalBehaviour* RobotController::getGlobalBehaviour()
{
  return myGlobalBehaviour;
}

EndEffectorBehavior* RobotController::getEndEffectorBehavior()
{
  return myEndEffectorBehavior;
}

void RobotController::setObstacleHandler (ObstaclesHandler* handler)
{
  for (uint i=0; i<myAvoidanceBehaviours.size(); i++ )
    myAvoidanceBehaviours[i]->setObstaclesHandler(handler);
}

void RobotController::ResetConfiguration(cVector newConf)
{
  myTransformhandler->setConfiguration(newConf);
  rvizPublisher->PublishJointState(newConf);
  torqueVector zeroSpeed;
  zeroSpeed.setZero();
  myTLC->setSpeedVector(zeroSpeed);
  btVector3 EEPos = myTransformhandler->getOrigins()[6];
  std::cout<<"Start Position: "<<EEPos.getX()<<" "<<EEPos.getY()<<" "<<EEPos.getZ()<<std::endl;
}

void RobotController::publishEEpos(btVector3 pos)
{
  kunz_brock_planner::EEpos sendData;
  sendData.header.stamp = ros::Time::now();
  sendData.x =  pos.getX();
  sendData.y =  pos.getY();
  sendData.z =  pos.getZ();
  if (ros::ok())
    EndEffectorPosPublisher.publish(sendData);
}

/**
   @brief Static "dummy function" for pthread_create
*/
void* RobotController::_run(void* arg) {
  reinterpret_cast<RobotController*>(arg)->run();
  return 0;
}

void RobotController::run()
{
  ros::Rate loop_rate(CONTROLLER_FREQ);
  ros::Time start_time;
  int k=0;
  while(true)
  {
    start_time = ros::Time::now();
    torqueVector newGamma = myTLC->getTorque();
    cVector currentConf = myTransformhandler->getConfiguration();
    cVector newConfiguration = myTLC->convertGammaToConfiguration(newGamma, currentConf);
    myTransformhandler->setConfiguration(newConfiguration);

    if(! (node_handler->ok()))
      pthread_exit(0);
    rvizPublisher->PublishJointState(newConfiguration);
    btVector3 EEposition = myTransformhandler->getOrigins()[myTransformhandler->getOrigins().size()-1];
    publishEEpos(EEposition);

    if (k%40==0)
      cout<<"RobotController::run: Control cycle duration "<<ros::Time::now() - start_time<<" sec"<<endl;
    k++;

    ros::spinOnce();
    loop_rate.sleep();
  }
}

Transformhandler* RobotController::getTransformhandler()
{
  return myTransformhandler;
}




