#include "SchunkTransformhandler.h"
#include <tf/tf.h>
#include <Eigen/Geometry>

SchunkTransformhandler::SchunkTransformhandler() {
  pthread_mutex_lock(&JacobianMutex);
  myJacobian.resize(6,7);
  myJacobian.setZero();
  pthread_mutex_unlock(&JacobianMutex);

  pthread_mutex_lock(&transformMutex);
  myTransforms.resize(7);
  myTransforms[0].setIdentity();
  myTransforms[1].setIdentity();
  myTransforms[2].setIdentity();
  myTransforms[3].setIdentity();
  myTransforms[4].setIdentity();
  myTransforms[5].setIdentity();
  myTransforms[6].setIdentity();

  rootToJointTransforms.resize(7);
  pthread_mutex_unlock(&transformMutex);

  InitJointInertia();
  InitJointLimits();
}

SchunkTransformhandler::~SchunkTransformhandler()
{}

void SchunkTransformhandler::updateTransforms() {
  //Wont update if transforms are already calculated for this configuration
  if (!shouldUpdate) return;


  tf::Transform joint1, joint2, joint3, joint4, joint5, joint6, joint7;
  tf::Quaternion YtoZ,ZtoY;
  ZtoY = tf::Quaternion(tf::Vector3(1,0,0), -SIMD_PI*.5) * tf::Quaternion(tf::Vector3(0,0,1), -SIMD_PI*.5);
  YtoZ = ZtoY.inverse();

  // Setting up the kinematic chain:
  joint1.setRotation(tf::Quaternion(tf::Vector3(0,0,-1), myConfiguration[0]));
  joint1.setOrigin(tf::Vector3(0, 0, 0.));

  joint2.setRotation(ZtoY * tf::Quaternion(tf::Vector3(0,0,1), myConfiguration[1]));
  joint2.setOrigin(tf::Vector3(0, 0., 0.3));

  joint3.setRotation(YtoZ *tf::Quaternion(tf::Vector3(0,0,-1), myConfiguration[2]));
  joint3.setOrigin(tf::Vector3(0.11, 0., 0.));

  joint4.setRotation(ZtoY *  tf::Quaternion(tf::Vector3(0,0,1), myConfiguration[3]));
  joint4.setOrigin(tf::Vector3(0., 0., 0.218));

  joint5.setRotation(YtoZ *tf::Quaternion(tf::Vector3(0,0,-1), myConfiguration[4]));
  joint5.setOrigin(tf::Vector3(0.1, 0, 0.));

  joint6.setRotation(ZtoY *  tf::Quaternion(tf::Vector3(0,0,1), myConfiguration[5]));
  joint6.setOrigin(tf::Vector3(0., 0., 0.1765));

  joint7.setRotation(YtoZ *tf::Quaternion(tf::Vector3(0,0,-1), myConfiguration[6]));
  joint7.setOrigin(tf::Vector3(0.171, 0, 0.));

  pthread_mutex_lock(&transformMutex);
  myTransforms[0] = joint1;
  myTransforms[1] = joint2;
  myTransforms[2] = joint3;
  myTransforms[3] = joint4;
  myTransforms[4] = joint5;
  myTransforms[5] = joint6;
  myTransforms[6] = joint7;

  rootToJointTransforms[0] = joint1;
  rootToJointTransforms[1] = joint1*joint2;
  rootToJointTransforms[2] = rootToJointTransforms[1]*joint3;
  rootToJointTransforms[3] = rootToJointTransforms[2]*joint4;
  rootToJointTransforms[4] = rootToJointTransforms[3]*joint5;
  rootToJointTransforms[5] = rootToJointTransforms[4]*joint6;
  rootToJointTransforms[6] = rootToJointTransforms[5]*joint7;
  pthread_mutex_unlock(&transformMutex);  

  shouldUpdate = false;

  jacobianCalculated = false;
}

/**
   @brief Calculates Jacobian of robot's end effector.
   Robot has 10 DOF (3 base + 7 arm), but Jacobian has only 9 columns:
   0 col - movement in forward/backward dir; 1 col - base rotation; other - arm.
*/
void SchunkTransformhandler::CalculateJacobian()
{
  if (jacobianCalculated)
          return;

  //Getting data about configuration
  std::vector<btVector3> rotAxes;
  std::vector<btVector3> rotOrigins;
  rotAxes = this->getAxes();
  rotOrigins = this->getOrigins();

  //Temporary local variables
  Eigen::Matrix<double, 6, 7> Jac_buf;
  Eigen::Matrix<double, 6, 1> Jac_column;
  Eigen::Vector3d Jv, endOrigin, prevOrigin, prevZ;
  Jv.setZero();
  Jac_column.setZero();
  endOrigin.setZero();
  prevOrigin.setZero();
  prevZ.setZero();

  //Position of end effector
  endOrigin = Eigen::Vector3d(rotOrigins.at(rotOrigins.size()-1).m_floats);

  //Arm joints
  for (int i = 0; i <= 6; ++i)
  {
    // Converting axes and origins from btVector3 to Eigen::Vector3d
    prevZ = Eigen::Vector3d(rotAxes[i].m_floats);
    prevOrigin = Eigen::Vector3d(rotOrigins[i].m_floats);
    // Revolute joint jacobian:
    // J_vi = z_i-1 x (o_n - o_n-1)
    Jv = prevZ.cross(endOrigin - prevOrigin);

    Jac_column.topRows(3) = Jv;
    Jac_column.bottomRows(3) = prevZ;
    Jac_buf.col(i+1) = Jac_column;
  }

  pthread_mutex_lock(&JacobianMutex);
  myJacobian = Jac_buf;
  pthread_mutex_unlock(&JacobianMutex);

  jacobianCalculated = true;
}

void SchunkTransformhandler::InitJointInertia()
{
  float mass[7] = {3.66, 4.26, 2.67, 2.57, 1.67, 1.51, 1.12}; //78
//geometric parameters of joints
  float diam[7] = {0.132, 0.132, 0.112, 0.112, 0.089, 0.089, 0.075};
  float h[7] = {0.18, 0.187, 0.118, 0.165, 0.097, 0.144, 0.092};
  double r[7] = { h[0]/2, 0, h[2]/2, 0,h[4]/2, 0, h[6]/2}; //distance from rotation point to CM

  Eigen::Matrix3d inertia_tmp;
  Eigen::Vector3d vectToCM_tmp;
  SpatialInertiaMatrix6d tmp;
  inertia_tmp.setIdentity();
  vectToCM_tmp[0] = vectToCM_tmp[1] = 0;   //translation only along Z-axis

  //other joints
  for (int i=0; i<7; i++)
  {
    inertia_tmp.setIdentity();
    inertia_tmp(0,0) = mass[i]*(3*diam[i-1] * diam[i-1] /4 + h[i-1]*h[i-1])/12;
    inertia_tmp(1,1) = mass[i]*(3*diam[i-1] * diam[i-1] /4 + h[i-1]*h[i-1])/12;
    inertia_tmp(2,2) = mass[i]*(diam[i-1] * diam[i-1] /4)/2;
    vectToCM_tmp[2] = r[i];
    tmp.setFromCMInertia(inertia_tmp,vectToCM_tmp,mass[i]);
    JointInertia.push_back(tmp);
  }
}

void SchunkTransformhandler::InitJointLimits()
{
  myJointLimits.setZero(2,7);
  myJointLimits.col(0) << -M_PI*2, M_PI*2; //Joint #1
  myJointLimits.col(1) << -123./180*M_PI, 123./180*M_PI; //Joint #2
  myJointLimits.col(2) << -M_PI*2, M_PI*2;//Joint #3
  myJointLimits.col(3) << -125./180*M_PI, 125./180*M_PI; //Joint #4
  myJointLimits.col(4) << -M_PI*2, M_PI*2; //Joint #5
  myJointLimits.col(5) << -125.5/180*M_PI, 125.5/180*M_PI; //Joint #6
  myJointLimits.col(6) << -M_PI*2, M_PI*2; //Joint #7


  jointIsLimited.setZero(7);
  jointIsLimited[0] = true;
  jointIsLimited[1] = true;
  jointIsLimited[2] = true;
  jointIsLimited[3] = true;
  jointIsLimited[4] = true;
  jointIsLimited[5] = true;
  jointIsLimited[6] = false;
}

/* @brief Prints debugging data*/
void SchunkTransformhandler::printData()
{
	std::vector<btVector3> rotAxes;
	std::vector<btVector3> rotOrigins;
	rotAxes = getAxes();
	rotOrigins = getOrigins();

	for (int i=0; i<7; i++)
	{
		std::cout<<std::setiosflags(std::ios::fixed) << std::setprecision(2)<<
				"Axis "<<i<<": "<<Eigen::Vector3d(rotAxes[i].m_floats).transpose()<<std::endl;
	}
	for (int i=0; i<7; i++)
	{
		std::cout<<"Origin "<<i<<": "<<Eigen::Vector3d(rotOrigins[i].m_floats).transpose()<<std::endl;
	}

	 Eigen::VectorXd q_eigen;
	 cVector q= myConfiguration;
	 q_eigen.resize(10);
	 for (int i=0;i<10;i++)
		 q_eigen [i]= q[i];
				  ;
	std::cout<<std::setiosflags(std::ios::fixed) << std::setprecision(2)<<"Jacobian: "<<std::endl<<myJacobian<<std::endl;
	//std::cout<<"Force: "<<F.transpose()<<std::endl;
	//std::cout<<"myGamma size: "<<myGamma.transpose()<<std::endl;
	std::cout<<"q: "<<q_eigen.transpose()<<std::endl;
}

///**
//   @brief Gets the robot base and manipulator joint origins in the bullet
//   world frame.
//*/
//std::vector<btVector3> SchunkTransformhandler::getOrigins()
//{
//    std::vector<btVector3> ret;
//    std::vector<btTransform>::iterator it;
//    btTransform transform;
//
//    //updateTransforms();
//    pthread_mutex_lock(&transformMutex);
//
//    transform.setIdentity();
//    transform.setOrigin(btVector3(0,0,0));
//
//    for (it = myTransforms.begin(); it < myTransforms.end(); ++it)
//    {
//            transform *= *it;
//            ret.push_back(transform.getOrigin());
//    }
//
//    pthread_mutex_unlock(&transformMutex);
//    return ret;
//}

///**
//   @brief Gets the robot base and joint rotational axes in the bullet
//   world frame. Base rotational axis is always parallell to the gravity
//   vector, while the joint rotation vectors could be "anywhere"
//   in the world frame.
//*/
std::vector<btVector3> SchunkTransformhandler::getAxes() {
  std::vector<btVector3> ret;
  btVector3 k(0., 0., 1.);
  btVector3 f(0., 0., -1.);

  pthread_mutex_lock(&transformMutex);
  for (uint i=0;i<rootToJointTransforms.size();i++)
    if (i%2==0)
      ret.push_back(rootToJointTransforms[i].getBasis()*f);
    else
      ret.push_back(rootToJointTransforms[i].getBasis()*k);
   pthread_mutex_unlock(&transformMutex);

  return ret;
}

