/// \brief Spatial vector algebra classes
/// 
/// This file includes spatial vector algebra classes intended for the
/// implementation of Featherstone's articulated rigid body dynamics method.
/// These classes are used for the robot's dynamics calculation.
//-------------------------------------------------------------------------------

#ifndef _SPATIAL_H_INCLUDED_
#define _SPATIAL_H_INCLUDED_

//-------------------------------------------------------------------------------
#include <btBulletDynamicsCommon.h>
#include <Eigen/Core>

//-------------------------------------------------------------------------------

/**
   @brief 6D spatial vector.
   This class is both for spatial motion vectors and spatial force vectors.
*/

class SpatialVector6d
{
public:
	Eigen::Vector3d a, b;

	///@name Constructors
	SpatialVector6d() {}
	SpatialVector6d( const Eigen::Vector3d &p1, const Eigen::Vector3d &p2 ) { Set(p1,p2); }
	SpatialVector6d( float a1, float a2, float a3, float b1, float b2, float b3 ) { Set(a1,a2,a3,b1,b2,b3); }
	SpatialVector6d( const SpatialVector6d &v ) { a=v.a; b=v.b; }

	///@name Initialization methods
	void Set( const Eigen::Vector3d &p1, const Eigen::Vector3d &p2 ) { a = p1; b = p2; }
	void Set( float a1, float a2, float a3, float b1, float b2, float b3 ) { a[0]=a1;a[1]=a2;a[2]=a3; b[0]=b1;b[1]=b2;b[2]=b3; }
	void setZero() { a.setZero(); b.setZero(); }

	///@name Unary operators
	SpatialVector6d operator-() const { return SpatialVector6d(-a,-b); }

	///@name Binary operators
	SpatialVector6d	operator + ( const SpatialVector6d &s ) const { return SpatialVector6d(a+s.a, b+s.b); }
	SpatialVector6d	operator - ( const SpatialVector6d &s ) const { return SpatialVector6d(a-s.a, b-s.b); }
	SpatialVector6d	operator * ( float t ) const { return SpatialVector6d( a*t, b*t ); }

	/*
	   *@brief Scalar product of two vectors.
	   * Note that one of the vectors should be motion vector ant the other should be a force vector.
	   * Otherwise, scalar product is not defined in spatial vector algebra.

	 */
	double dot ( const SpatialVector6d &s ) const { return a.dot(s.a) + b.dot(s.b); }

	///@name Assignment operators
	SpatialVector6d & operator =  ( const SpatialVector6d &v ) { a=v.a; b=v.b; return *this; }
	SpatialVector6d & operator += ( const SpatialVector6d &s ) { a+=s.a; b+=s.b; return *this;}
	SpatialVector6d & operator -= ( const SpatialVector6d &s ) { a-=s.a; b-=s.b; return *this;}
	SpatialVector6d & operator *= ( float t ) { a*=t; b*=t; return *this;}

	///@name Access operators
	double& operator [] (int i)
	{
	//	double& ret = 0;
		if ((i>=0) && (i<3))
			return a[i];
		else
			if ((i>=3) && (i<6))
				return b[i-3];
			else
			{
				std::cout<<"SpatialVector6d::operator[]: ERROR: Input value must be between 0 and 5!"<<std::endl;
				exit(1);
			}
	}

};


//-------------------------------------------------------------------------------

/**
   @brief 6D spatial matrix for coordinate transforms. This is a special case for 6D spatial matrix,
   where the matrix represents a coordinate transformation. In this case, instead of keeping a full
   6x6 matrix values, we can keep a 3x3 matrix for rotation, and a 3D point for translation.
*/
class SpatialTrans6d
{
public:

	// |    R     0 |
	// | -r x R   R |


	/**   @brief Rotation matrix */
	Eigen::Matrix3d R;
	/**   @brief Translation vector */
	Eigen::Vector3d r;

	///@name Constructors
	SpatialTrans6d() {}
	SpatialTrans6d( const SpatialTrans6d &mat ) { R=mat.R; r=mat.r; }
	SpatialTrans6d( const Eigen::Matrix3d &_R, const Eigen::Vector3d &_r ) { R=_R; r=_r; }

	///@name Initialization methods
	void Set( const Eigen::Matrix3d &_R, const Eigen::Vector3d &_r ) { R=_R; r=_r; }
	void SetIdentity() { R.setIdentity();r.setZero(); }

	///@name Binary operators
	SpatialVector6d operator * ( const SpatialVector6d &p ) const
	{
		return SpatialVector6d( R*p.a,   R*(p.b - r.cross(p.a)));
	}

	SpatialTrans6d operator * ( const SpatialTrans6d &mat ) const
			{ return SpatialTrans6d( R*mat.R, mat.r + mat.R.transpose()*r ); }

	///@name Assignment operators
	SpatialTrans6d & operator =  ( const SpatialTrans6d &mat ) { R=mat.R; r=mat.r; return *this; }
	SpatialTrans6d & operator *= ( const SpatialTrans6d &mat )
			{ *this = *this * mat; return *this;}


	/*@brief Performs operation X.transpose()*F (F - spatial vector).
	 * This function is implemented separately, as short representation of transforms
	 * doesn't allow explicit transposition */
	SpatialVector6d multiplyTransposedXWithVector (const SpatialVector6d &p ) const
	{
		Eigen::Matrix3d S;//skew-symmetric matrix
		S<<0, -r[2], r[1], r[2], 0, -r[0], -r[1], r[0], 0;
		Eigen::Matrix3d R_t;
		R_t = R.transpose();
		return SpatialVector6d(R_t*p.a + S*R_t*p.b,  R_t*p.b);
	}

};

//----------------------------------------------------
/**
   @brief Class for representation of Spatial inertia matrix and all possible operations
*/
class SpatialInertiaMatrix6d
{
public:

	// |    I_o     			S(h) 		|
	// | 	S(h).transpose   	m*Identity 	|

	/**   @brief Inertia tensor across rotation point */
	Eigen::Matrix3d I_o;
	/**   @brief h = mass*c, where c - vector from rotation point to CM (h - first moment of mass) */
	Eigen::Vector3d h;
	/**   @brief mass */
	double mass;

	///@name Constructors
	SpatialInertiaMatrix6d() {}
	SpatialInertiaMatrix6d( const SpatialInertiaMatrix6d &mat ) { I_o=mat.I_o; h=mat.h; mass = mat.mass;}
	SpatialInertiaMatrix6d( const Eigen::Matrix3d &_I_o, const Eigen::Vector3d &_h, const double &_mass)
		{ I_o=_I_o; h=_h; mass = _mass;}

	///@name Initialization methods
	void Set( const Eigen::Matrix3d &_I_o, const Eigen::Vector3d &_h, const double &_mass )
		{I_o=_I_o; h=_h; mass = _mass;}
	void SetIdentity() { I_o.setIdentity();h.setZero();mass =1; }

	/**
	   @brief Forms spatial inertia matrix from usual CM parameters
	   @param _I_cm - Center of mass Inertia tensor
	   @param _c - vector from CM to rotation point
	   @param _mass - mass
	*/
	void setFromCMInertia(const Eigen::Matrix3d &_I_cm, const Eigen::Vector3d &_c, const double &_mass)
	{
		mass = _mass;
		h=_c * mass;
		Eigen::Matrix3d S;//skew-symmetric matrix
		S<<0, -_c[2], _c[1], _c[2], 0, -_c[0], -_c[1], _c[0], 0;
		I_o=_I_cm + mass * S *S.transpose();
	}

	///@name Binary operators
	SpatialVector6d operator * ( const SpatialVector6d &p ) const
	{
		return SpatialVector6d(I_o*p.a + h.cross(p.b), mass*p.b - h.cross(p.a));
	}

	SpatialInertiaMatrix6d operator + ( const SpatialInertiaMatrix6d &mat ) const
	{
		return SpatialInertiaMatrix6d( I_o + mat.I_o, h + mat.h, mass + mat.mass );
	}
	SpatialInertiaMatrix6d operator - ( const SpatialInertiaMatrix6d &mat ) const
	{
		return SpatialInertiaMatrix6d( I_o - mat.I_o, h - mat.h, mass - mat.mass );
	}

	/**
	   @brief Translates Inertia to another coordinate frame: I_new = X.transpose * I_old * X
	   @param X - Spatial matrix for coordinate transform
	*/
	SpatialInertiaMatrix6d TranslateInertiaToOtherFrame (const SpatialTrans6d &X)
	{
		Eigen::Matrix3d Sp, SRh, SRhmp;	//skew-symmetric matrices
		Eigen::Vector3d Rh,Rhmp;		//intermediate vectors

		Rh = X.R.transpose()*h;
		Rhmp = Rh + mass*X.r;

		Sp<<0, -X.r[2], X.r[1], X.r[2], 0, -X.r[0], -X.r[1], X.r[0], 0;
		SRh<<0, -Rh[2], Rh[1], Rh[2], 0, -Rh[0], -Rh[1], Rh[0], 0;
		SRhmp<<0, -Rhmp[2], Rhmp[1], Rhmp[2], 0, -Rhmp[0], -Rhmp[1], Rhmp[0], 0;

		return SpatialInertiaMatrix6d( X.R.transpose()*I_o*X.R - Sp*SRh - SRhmp*Sp,
								Rhmp,
								mass);
	}

	///@name Assignment operators
	SpatialInertiaMatrix6d & operator =  ( const SpatialInertiaMatrix6d &mat ) { I_o = mat.I_o; h = mat.h; mass = mat.mass; return *this;}
	SpatialInertiaMatrix6d & operator += ( const SpatialInertiaMatrix6d &mat ) { I_o += mat.I_o; h += mat.h; mass += mat.mass; return *this;}
	SpatialInertiaMatrix6d & operator -= ( const SpatialInertiaMatrix6d &mat ) { I_o -= mat.I_o; h -= mat.h; mass -= mat.mass; return *this;}

};


//-------------------------------------------------------------------------------


#endif
