#include "TaskLevelController.h"


#include "Spatial.h"
#include <Eigen/Dense>
#include <iomanip>
#include <iostream>

using namespace std;

/**
   @brief Constructor
   @param robot Reference to the robot model that receive the torques
   @param numDofs Degrees of freedom
*/
TaskLevelController::TaskLevelController(RobotController* const parent, int numDof)
  : myRobot(parent)
{
  myGamma.resize(numDof);
  myGamma.setZero();
  first = true;

  H.resize(numDof,numDof);
  H.setZero();
  N.resize(numDof,numDof);
  N.setIdentity();
  H_inverse.resize(numDof,numDof);
  H_inverse.setZero();
  NumberOfDOF = numDof;
}

TaskLevelController::~TaskLevelController()
{
  std::list<Behaviour*>::iterator it;
  while (myBehaviours.size()>0)
  {
    Behaviour* beh = myBehaviours.back();
    myBehaviours.pop_back();
    delete beh;
  }
}


/**
   @brief Takes current configuration and calculates operational-space and joint-space
   * inertia matrix and its inverse with Composite-Rigid-Body algorithm.
   * See R.Featherstone "Rigid Body Dynamics Algorithms"
   * 	   or "Springer Handbook of Robotics" (Chapter 2) for the algorithm explanation
*/
void TaskLevelController::computeInertiaMatrices()
{
	std::vector<btTransform> Transforms = myRobot->getTransformhandler()->getTransforms();
	Eigen::Matrix3d transform;
	transform.setIdentity();
	JointInertia_C = *myRobot->getTransformhandler()->getJointInertia();
	std::vector<SpatialTrans6d> SpatialTransforms;
	SpatialTrans6d X;

	for(int i=JointInertia_C.size()-2;i>=0;i--)
	{
		transform.col(0) = Eigen::Vector3d((Transforms[i+1].getBasis()[0]).m_floats);
		transform.col(1) = Eigen::Vector3d((Transforms[i+1].getBasis()[1]).m_floats);
		transform.col(2) = Eigen::Vector3d((Transforms[i+1].getBasis()[2]).m_floats);
		X.Set(transform, Eigen::Vector3d((Transforms[i+1].getOrigin().m_floats)));
		SpatialTransforms.push_back(X);
		JointInertia_C[i] = JointInertia_C[i] + JointInertia_C[i+1].TranslateInertiaToOtherFrame(X);
	}

	//matrix representing the free modes of the joints ("Ф" in Springer book of Robotics)
	//S1 (0 0 1 0 0 0) - rotation around axis z

	SpatialVector6d S1, F1;
	S1.setZero();
	S1[2] = 1;

	H.setZero();
	for (int i = JointInertia_C.size()-1; i >= 0; i--)
	{
          F1 = JointInertia_C[i]*S1;

          H(i,i) = S1.dot(F1);
          if (i>0)
          {
            for (int j=i-1; j >=0; j--)
            {
              F1 = SpatialTransforms[SpatialTransforms.size()-1 - j].multiplyTransposedXWithVector(F1);

              H(i,j) = S1.dot(F1);
              H(j,i) = H(i,j);
            }
	  }
	}

	////Computation of inverse and operational-space matrices
	// only if H has reasonable values
	if ((H.trace()>0))
	{
		H_inverse = H.inverse();
	}
	else
	{
		H_inverse.setZero();
	}

//        //Checking that the matrix is positive-defined
//        //(Only for test reasons)
//        Eigen::VectorXcd eivals = H.eigenvalues();
//
//        for(int i = 0; i< eivals.rows()-1;i++)
//        {
//                if ( std::real(eivals[i]) <= 0 )
//                {
//                        std::cout<<"DEBUG:___ H: "<<std::endl<<H<<std::endl<<
//                                        Eigen::Vector3d(Transforms[3].getBasis()[1].m_floats).transpose()<<" inert "<<JointInertia_C[1].h.transpose()<<
//                                        std::endl<<"Eigen values of H:"<<std::endl<<eivals<<std::endl<<
//                                "TaskLevelController::computeInertiaMatrices: ERROR: Eigen values of inertia matrix are negative"<<std::endl;
//                        exit(1);
//                }
//                if ( std::imag(eivals[i]) != 0 )
//                {
//                        std::cout<<"DEBUG:___H: "<<std::endl<<H<<std::endl<<
//                                        Eigen::Vector3d(Transforms[3].getBasis()[1].m_floats).transpose()<<" inert "<<JointInertia_C[1].h.transpose()<<
//                                        std::endl<<"Eigen values of H:"<<std::endl<<eivals<<std::endl
//                                <<"TaskLevelController::computeInertiaMatrices: ERROR: Eigen values of inertia matrix are complex"<<std::endl;
//                        exit(1);
//                }
//        }

}

/**
   @brief Gathers force output from all subordinate behaviours and combines
   them to a single torque vector. 

   @return Torque vector
 */
torqueVector TaskLevelController::getTorque()
{
        std::list<Behaviour*>::iterator it;
	Eigen::Vector3d F;
	torqueVector gamma, BehaviorTorque;

	computeInertiaMatrices();

	myGamma.setZero();
	gamma.setZero();
	BehaviorTorque.setZero();

	N.setIdentity();


	for (it = myBehaviours.begin(); it != myBehaviours.end(); it++)
	{
		BehaviorTorque = (*it)->getTorque();
		gamma += BehaviorTorque;
		N = (*it)->getResultingNullspace();
		//cout<<gamma.transpose()<<endl;

	}
	//cout<<endl;
	std::cout<<std::setiosflags(std::ios::fixed) << std::setprecision(3);
	myGamma = gamma;
	return myGamma;
}

/**
   @brief Adds a behaviour to the priority based list. The first behaviour added
   has highest priority.
   @param newBehaviour
*/
void TaskLevelController::addBehaviour(Behaviour* newBehaviour)
{
  myBehaviours.push_back(newBehaviour);
}

/**
   @brief Removes all behaviours
*/
void TaskLevelController::removeBehaviours() {
  myBehaviours.clear();
}

/**
   @brief Converts a set of joint torques to a new configuration.
   	   NOTE: This is the simulation of specific robot, Therefore this function is Robot-dependent.
   Integrates the following system of equations:
    	p_dot_new=p_dot_prev + H_inverse * gamma*time;
	p - set of contollable variables (for base - movement forward + rotation). For SeekurJr+LWA3 dim(p)=9.
	q - set of configuration variables. For SeekurJr+LWA3 dim(q)=10.
	Going from p to q is performed with the following equations:
	for arm jopints:    	 p_new = p_prev + p_dot_prev*time;
	for the mobile base:	x=q[0] = x_prev + sin(p[1])*p_dot_prev[0]*time;
							z=q[1] = z_prev + cos(p[1])*p_dot_prev[0]*time;
							theta = q[2] = p_prev[1] + p_dot_prev[1]*time;
*/
cVector TaskLevelController::convertGammaToConfiguration(torqueVector gamma, cVector& currentConfig)
{
  cVector new_q, new_q_buf;
  torqueVector delta, new_p, new_p_dot, new_p_dot_dot;

  ptime now = microsec_clock::local_time();
  if (first)
  {
    first = false;
    new_q = currentConfig;
    p_dot_prev.setZero();
  }
  else
  {
    double timeDiff = (now-then).total_microseconds()/1000000.0;

    timeDiff = 0.025; //REMOVE THIS FOR VARIABLE TIME STEP


	cVector q = currentConfig;

	new_p_dot_dot = torqueVector(H_inverse*gamma);

	//Limiting max angular speed of the joints according to motor parameters
	  if (fabs(new_p_dot_dot[6])>PRL60_ACCELERATION_MAX)
		  new_p_dot_dot[6]=PRL60_ACCELERATION_MAX*signof(new_p_dot_dot[6]);
	  if (fabs(new_p_dot_dot[5])>PRL80_ACCELERATION_MAX)
		  new_p_dot_dot[5]=PRL80_ACCELERATION_MAX*signof(new_p_dot_dot[5]);
	  if (fabs(new_p_dot_dot[4])>PRL80_ACCELERATION_MAX)
		  new_p_dot_dot[4]=PRL80_ACCELERATION_MAX*signof(new_p_dot_dot[4]);
	  if (fabs(new_p_dot_dot[3])>PRL100_ACCELERATION_MAX)
		  new_p_dot_dot[3]=PRL100_ACCELERATION_MAX*signof(new_p_dot_dot[3]);
	  if (fabs(new_p_dot_dot[2])>PRL100_ACCELERATION_MAX)
		  new_p_dot_dot[2]=PRL100_ACCELERATION_MAX*signof(new_p_dot_dot[2]);
	  if (fabs(new_p_dot_dot[1])>PRL120_ACCELERATION_MAX)
		  new_p_dot_dot[1]=PRL120_ACCELERATION_MAX*signof(new_p_dot_dot[1]);
	  if (fabs(new_p_dot_dot[2])>PRL120_ACCELERATION_MAX)
		  new_p_dot_dot[0]=PRL120_ACCELERATION_MAX*signof(new_p_dot_dot[0]);


	  new_p_dot = p_dot_prev + new_p_dot_dot * timeDiff;

	//Limiting max angular speed of the joints according to motor parameters
	  if (fabs(new_p_dot[6])>PRL60_SPEED_MAX)
		  new_p_dot[6]=PRL60_SPEED_MAX*signof(new_p_dot[6]);
	  if (fabs(new_p_dot[5])>PRL80_SPEED_MAX)
		  new_p_dot[5]=PRL80_SPEED_MAX*signof(new_p_dot[5]);
	  if (fabs(new_p_dot[4])>PRL80_SPEED_MAX)
		  new_p_dot[4]=PRL80_SPEED_MAX*signof(new_p_dot[4]);
	  if (fabs(new_p_dot[3])>PRL100_SPEED_MAX)
		  new_p_dot[3]=PRL100_SPEED_MAX*signof(new_p_dot[3]);
	  if (fabs(new_p_dot[2])>PRL100_SPEED_MAX)
		  new_p_dot[2]=PRL100_SPEED_MAX*signof(new_p_dot[2]);
	  if (fabs(new_p_dot[1])>PRL120_SPEED_MAX)
		  new_p_dot[1]=PRL120_SPEED_MAX*signof(new_p_dot[1]);
	  if (fabs(new_p_dot[0])>PRL120_SPEED_MAX)
		  new_p_dot[0]=PRL120_SPEED_MAX*signof(new_p_dot[0]);

//	  for (int i=0; i<H.cols(); i++)
//		  if (fabs(new_p_dot[i])<0.005)
//			  new_p_dot[i] = 0;

	  for (int i = 0; i<H.cols(); i++)
	  {
	    new_q[i] = q[i] + p_dot_prev[i]*timeDiff;
	  }

	  p_dot_prev = new_p_dot;
  }

  then = now;
  return new_q;
}

const Eigen::MatrixXd& TaskLevelController::getInertiaMatrixInverse()
{
	return H_inverse;
}

const Eigen::MatrixXd& TaskLevelController::getInertiaMatrix()
{
	return H;
}
const Eigen::MatrixXd& TaskLevelController::getNullSpaceMatrix()
{	return N;}

const torqueVector& TaskLevelController::getSpeedVector()
{	return p_dot_prev;}

void TaskLevelController::setSpeedVector(const torqueVector& new_p_dot)
{	p_dot_prev = new_p_dot;}

RobotController* TaskLevelController::getRobot()
{	return myRobot;}

int TaskLevelController::getNumberOfDOF()
{	return NumberOfDOF;}

//bool TaskLevelController::getEndEffectorTaskSuccess()
//{
//	EndEffectorTaskBehaviour* behavior=OwningMilestone->getEndEffectorTaskBehaviour();
//	if (behavior)
//		return behavior->getSuccessStatus();
//	else
//	{
//		std::cout<<"DEBUG:___TaskLevelController::getEndEffectorTaskSuccess: WARNING: trying to access nonexisting behavior"<<std::endl;
//		return true;
//	}
//}

//bool TaskLevelController::getGlobalTaskActive()
//{
//	GlobalBehaviour* behavior=OwningMilestone->getGlobalTaskBehaviour();
//	if (behavior)
//		return behavior->getActive();
//	else
//	{
//		return false;
//	}
//}


