#ifndef __TASKLEVELCONTROLLER_H__
#define __TASKLEVELCONTROLLER_H__

#include <boost/date_time/posix_time/posix_time.hpp>
#include <list>
#include <math.h>


#include "RobotController.h"
#include "Transformhandler.h"
#include "Behaviours/Behaviour.h"
#include "matrixDefinition.h"

#include "Spatial.h"

#define PRL60_TORQUE_MAX 4.5
#define PRL80_TORQUE_MAX 20.7
#define PRL100_TORQUE_MAX 81.5
#define PRL120_TORQUE_MAX 216

#define PRL60_ACCELERATION_MAX 3.49
#define PRL80_ACCELERATION_MAX 1.75
#define PRL100_ACCELERATION_MAX 1.66
#define PRL120_ACCELERATION_MAX 1.75

#define PRL60_SPEED_MAX 0.87
#define PRL80_SPEED_MAX 0.44
#define PRL100_SPEED_MAX 0.42
#define PRL120_SPEED_MAX 0.44

class RobotController;
class Behaviour;

using namespace boost::posix_time;

/**
   @brief Operational space controller. Provides a torque vector based on
   a priority-list of force-generating behaviours.
*/
class TaskLevelController {
 public:
  TaskLevelController(RobotController* const,int);
  ~TaskLevelController();
  void addBehaviour(Behaviour* newBehaviour);
  void removeBehaviours();
  torqueVector getTorque();
  cVector convertForceToTorque(cVector);
  cVector convertGammaToConfiguration(torqueVector gamma, cVector& currentConfig);
  EIGEN_MAKE_ALIGNED_OPERATOR_NEW //Aligning bytes in matrices (libEigen tricks)
  void computeInertiaMatrices();

  const Eigen::MatrixXd& getInertiaMatrixInverse();
  const Eigen::MatrixXd& getInertiaMatrix();
  const Eigen::MatrixXd& getNullSpaceMatrix();
  RobotController* getRobot();
  int getNumberOfDOF();
  bool getEndEffectorTaskSuccess();
  bool getGlobalTaskActive();
  const torqueVector& getSpeedVector();
  void setSpeedVector(const torqueVector& new_p_dot);
 protected:
  inline int signof(double a) { return (a != 0) ? (a<0 ? -1 : 1) : 0; }
  RobotController* const myRobot;
  torqueVector myGamma;
  /** @brief List of force-generating behaviours. myBehaviours[0] has the highest priority. */
  std::list<Behaviour*> myBehaviours;
 private:
  bool first;
  ptime then;  

  int NumberOfDOF;

  /** @brief Spatial inertia matrices used in Composite-Rigid-Body algorithm (depend on q)*/
  std::vector<SpatialInertiaMatrix6d> JointInertia_C;
  /** @brief Joint-space inertia matrix */
  Eigen::MatrixXd H;
  Eigen::MatrixXd H_inverse;
  /** @brief Null-space matrix*/
  Eigen::MatrixXd N;


 torqueVector p_dot_prev;
};

#endif /*__TASKLEVELCONTROLLER_H__*/
