#include "Transformhandler.h"

#include <btBulletDynamicsCommon.h>

Transformhandler::Transformhandler()
{
    shouldUpdate = true;
    pthread_mutex_init(&transformMutex, 0);
    pthread_mutex_init(&configMutex, 0);
    pthread_mutex_init(&JacobianMutex, 0);
};

std::vector<btTransform> Transformhandler::getTransforms() {
  if (shouldUpdate) {
    updateTransforms();
    shouldUpdate = false;
  }

  pthread_mutex_lock(&transformMutex);
  std::vector<btTransform> ret = myTransforms;
  pthread_mutex_unlock(&transformMutex);

  return ret;
}

std::vector<btTransform> Transformhandler::getRootToJointTransforms() {
  if (shouldUpdate) {
    updateTransforms();
    shouldUpdate = false;
  }

  pthread_mutex_lock(&transformMutex);
  std::vector<btTransform> ret = rootToJointTransforms;
  pthread_mutex_unlock(&transformMutex);

  return ret;
}

btTransform Transformhandler::getRootToJointTransform(uint i){
  if (shouldUpdate) {
    updateTransforms();
    shouldUpdate = false;
  }
  btTransform a;
  if (i<rootToJointTransforms.size())
  {
    pthread_mutex_lock(&transformMutex);
    a = rootToJointTransforms[i];
    pthread_mutex_unlock(&transformMutex);
  }
  else
  {
    std::cout<<"WARNING: Transformhandler::getRootToJointTransform: Trying to access non-existing joint."<<std::endl;
    a.setIdentity();
  }
  return a;
}

/**
   @brief Returns the Jacobian of current configuration
   (operating point - End effector)
*/
const Eigen::MatrixXd  Transformhandler::getJacobian()
{
	 Eigen::MatrixXd Jac_buf;
	 CalculateJacobian();
	pthread_mutex_lock(&JacobianMutex);
	Jac_buf = myJacobian;
	pthread_mutex_unlock(&JacobianMutex);
	return Jac_buf;
}

/**
   @brief Calculates the origins of the coordinate frames (in world coordinates)
   @return vector of positions
*/
std::vector<btVector3> Transformhandler::getOrigins() {
  std::vector<btVector3> ret;

  pthread_mutex_lock(&transformMutex);
  for (uint i=0;i<rootToJointTransforms.size();i++)
    ret.push_back(rootToJointTransforms[i].getOrigin());
   pthread_mutex_unlock(&transformMutex);
  
  return ret;
}

cVector Transformhandler::getConfiguration() {
  pthread_mutex_lock(&configMutex);
  cVector ret = myConfiguration;
  pthread_mutex_unlock(&configMutex);

  return ret;
}

void Transformhandler::setConfiguration(const cVector& config)
{
	pthread_mutex_lock(&configMutex);
	  myConfiguration = config;
	  pthread_mutex_unlock(&configMutex);
	  shouldUpdate = true;
	  updateTransforms();

}

void Transformhandler::setTransforms(const std::vector<btTransform>& transforms) {
  pthread_mutex_lock(&transformMutex);
  myTransforms = transforms;
  pthread_mutex_unlock(&transformMutex);
}

/**
   @brief Calculates a vector of the z-axes (in world coordinates) of the transforms.
   NOTE: This should be overridden if an inheriting class does
   not use the z axis as rotational axis.
*/
std::vector<btVector3> Transformhandler::getAxes() {
  std::vector<btVector3> ret;
  btVector3 k(0., 0., 1.);

  pthread_mutex_lock(&transformMutex);
  for (uint i=0;i<rootToJointTransforms.size();i++)
    ret.push_back(rootToJointTransforms[i].getBasis()*k);
   pthread_mutex_unlock(&transformMutex);
  
  return ret;
}

/**
   @brief Calculates a vector of the joint coordinate frames (in world coordinates) of the transforms.
*/
std::vector<btMatrix3x3> Transformhandler::getBasises() {
  std::vector<btMatrix3x3> ret;

  pthread_mutex_lock(&transformMutex);
  for (uint i=0;i<rootToJointTransforms.size();i++)
    ret.push_back(rootToJointTransforms[i].getBasis());
   pthread_mutex_unlock(&transformMutex);
  return ret;
}

std::vector<SpatialInertiaMatrix6d>* Transformhandler::getJointInertia()
{
  if(JointInertia.size() == 0)
    std::cout << "Transformhandler::getJointInertia(): WARNING: Joint Inertia vector has not been initialized.\n";
  return &JointInertia;
}

void Transformhandler::getJointLimits(Eigen::MatrixXd& limits, Eigen::VectorXi& isLimitedArray )
{
  limits = myJointLimits;
  isLimitedArray = jointIsLimited;
}
